from __future__ import annotations


from cohdl._core._type_qualifier import (
    TypeQualifierBase,
    Signal,
)
from cohdl._core import (
    Bit,
    BitVector,
    Unsigned,
    true,
    false,
    Null,
    Full,
    static_assert,
    concurrent_context,
    expr_fn,
    AssignMode,
    always,
    is_primitive_type,
)

from cohdl._core import Array as CohdlArray

from ._core_utility import (
    Ref,
    Value,
    nop,
    Nonlocal,
    zeros,
    width,
    concat,
    to_bits,
    from_bits,
    count_bits,
    base_type,
    is_qualified,
    instance_check,
    as_pyeval,
)
from ._template import Template, template_arg
from ._assignable_type import AssignableType

from cohdl._core._intrinsic import _intrinsic


from ._context import Duration, SequentialContext, concurrent, at_end_of_context
from ._prefix import prefix, name, NamedQualifier


# a singleton only used internally in the std module
class _None:
    pass


@_intrinsic
def add_entity_port(entity, port, name: str | None = None):
    if name is None:
        name = port.name()
        assert name is not None, "cannot determine name of new port"

    assert name not in entity._info.ports, f"port '{name}' already exists"

    entity._info.add_port(name, port)
    return port


class _SerializedTemplateArgs:
    elem_type: type

    @_intrinsic
    def __init__(self, elem_type: type):
        assert not isinstance(
            elem_type, Serialized
        ), "Serialized types should not be nested"
        self.elem_type = elem_type

    @_intrinsic
    def __hash__(self):
        return id(self.elem_type)

    @_intrinsic
    def __eq__(self, other: _SerializedTemplateArgs):
        return self.elem_type is other.elem_type


class Serialized(Template[_SerializedTemplateArgs], AssignableType):
    _elemtype_: _SerializedTemplateArgs.elem_type

    @classmethod
    def from_raw(cls, raw: BitVector):
        return cls(raw, _use_as_raw=True)

    def _assign_(self, source, mode: AssignMode) -> None:
        if isinstance(source, Serialized):
            assert (
                self._elemtype_ is source._elemtype_
            ), "elem type of source does not match elem type of target"
            self._raw._assign_(source._raw, mode)
        else:
            assert self._elemtype_ is base_type(
                source
            ), "source type does not match elem type of target"
            self._raw._assign_(to_bits(source), mode)

    def __init__(self, raw=None, _qualifier_=Value, _use_as_raw=False):
        elem_type = self._elemtype_
        bit_count = count_bits(elem_type)

        if _use_as_raw:
            assert (
                bit_count == raw.width
            ), "the number serialized bits ({}) does not match required width ({}) of type '{}'".format(
                raw.width, bit_count, elem_type
            )
            self._raw = raw.bitvector
        else:
            if isinstance(raw, Serialized):
                assert (
                    elem_type is raw._elemtype_
                ), "elem type of source does not match elem type of target"
                self._raw = _qualifier_[elem_type](raw._raw)
            elif raw is Null or raw is Full:
                self._raw = _qualifier_[BitVector[bit_count]](BitVector[bit_count](raw))
            else:
                assert (
                    base_type(raw) is elem_type
                ), "source type does not match elem type of target"
                self._raw = to_bits(raw)

    def value(self, qualifier=Value):
        return from_bits[self._elemtype_](self._raw, qualifier)

    @property
    def ref(self):
        return self.value(qualifier=Ref)

    def bits(self):
        return self._raw


#
#
#


def as_readable_vector(*parts):
    total_width = sum(width(p) for p in parts)

    result = Signal[BitVector[total_width]]()

    @concurrent
    def logic():
        result.next = concat(*parts)

    return result


def as_writeable_vector(*parts, default=None):
    offset = 0
    slice_list = []

    for part in parts:
        if instance_check(part, Bit):
            slice_list.append(part, offset)
            offset += 1
        else:
            w = width(part)
            slice_list.append((part, slice(offset + w - 1, offset)))
            offset += w

    total_width = offset

    result = Signal[BitVector[total_width]](default)

    @concurrent
    def logic():
        for part, slice in slice_list:
            part <<= result[slice]

    return result


#
#
#


class _ArrayArgs:
    def __init__(self, args: tuple):
        assert (
            isinstance(args, tuple) and len(args) == 2
        ), "std.Array should be declared like this: std.Array[TYPE, ELEM_CNT]"

        elem_type, elem_cnt = args
        self.elem_type = elem_type
        self.elem_cnt = int(elem_cnt)

    def __hash__(self):
        return hash((self.elem_type, self.elem_cnt))

    def __eq__(self, other: _ArrayArgs):
        return self.elem_type is other.elem_type and self.elem_cnt == other.elem_cnt


@_intrinsic
def _check_array_defaults(defaults: list, elem_width: int):
    for nr, elem in enumerate(defaults):
        assert not isinstance(
            elem, TypeQualifierBase
        ), f"default value at index {nr} is not compile time constant"

        assert isinstance(
            elem, BitVector[elem_width]
        ), f"serialized type of default element {nr} ({type(elem)}) does not match type needed for array elements ({BitVector[elem_width]})"


class Array(Template[_ArrayArgs], AssignableType):
    _count_: _ArrayArgs.elem_cnt
    _elemtype_: _ArrayArgs.elem_type

    def _assign_(self, source, mode: AssignMode) -> None:
        if isinstance(source, (Array, list, CohdlArray)):
            assert self._count_ == len(source)

            for index in range(self._count_):
                self[index]._assign_(source[index], mode)
        elif isinstance(source, dict):
            for index, val in source.items():
                self[index]._assign_(val, mode)
        else:
            assert (
                source is Null or source is Full
            ), "invalid argument type for array assignment"

            for index in range(self._count_):
                self[index]._assign_(source, mode)

    def __init__(self, val=None, *, name=None, _qualifier_=Signal, attributes=None):
        elem_width = count_bits(self._elemtype_)

        if val is None or val is Null or val is Full:
            self._content = _qualifier_[
                CohdlArray[BitVector[elem_width], self._count_]
            ](val, name=name, attributes=attributes)
        else:
            assert isinstance(
                val, (list, tuple)
            ), "default argument must be a list or a tuple"
            assert (
                len(val) <= self._count_
            ), "to many default arguments for array ({} > {})".format(
                len(val), self._count_
            )

            default_as_elemtype = [
                (
                    val_elem
                    if isinstance(val_elem, self._elemtype_)
                    else Value[self._elemtype_](val_elem)
                )
                for val_elem in val
            ]

            default_as_bitvector = [
                to_bits(default_elem) for default_elem in default_as_elemtype
            ]

            _check_array_defaults(default_as_bitvector, elem_width)

            self._content = _qualifier_[
                CohdlArray[BitVector[elem_width], self._count_]
            ](default_as_bitvector, name=name, attributes=attributes)

    @_intrinsic
    def __len__(self):
        return self._count_

    def get_elem(self, index, qualifier=Ref):
        return from_bits[self._elemtype_](self._content[index], qualifier)

    def set_elem(self, index, value):
        self._content[index]._assign_(value, AssignMode.AUTO)

    def __getitem__(self, index):
        return self.get_elem(index, Ref)


#
#
#


@_intrinsic
def stringify(*args):
    return "".join(str(arg) for arg in args)


class DelayLine:
    def __init__(self, inp, delay: int, initial=_None, ctx: None = None):
        qualifier = NamedQualifier[Nonlocal[Signal], "delayline"][base_type(inp)]

        if initial is _None:
            self._steps = [
                inp,
                *[qualifier() for _ in range(delay)],
            ]
        else:
            self._steps = [
                inp,
                *[qualifier(initial) for _ in range(delay)],
            ]

        def delay_impl():
            for src, target in zip(self._steps, self._steps[1:]):
                target <<= src

        if ctx is not None:

            @ctx
            def process_delay():
                delay_impl()

        else:
            delay_impl()

    def __getitem__(self, delay: int):
        return self._steps[delay]

    def __len__(self):
        return len(self._steps)

    def __iter__(self):
        return iter(self._steps)

    def last(self):
        return self._steps[-1]


def delayed(inp, delay: int, initial=_None):
    return DelayLine(inp, delay, initial=initial).last()


def debounce(
    ctx: SequentialContext,
    inp: Signal[Bit],
    period: int | Duration,
    initial=False,
    allowed_delta=1e-9,
):
    if isinstance(period, Duration):
        period = period.count_periods(ctx.clk().period(), allowed_delta=allowed_delta)

    assert period >= 1, "debounce period to small"

    result = Signal[Bit](initial)
    counter = Signal[Unsigned.upto(period)](period // 2)

    @ctx
    def proc_debounce():
        if inp:
            if counter == period:
                result.next = True
            else:
                counter.next = counter + 1
        else:
            if counter == 0:
                result.next = False
            else:
                counter.next = counter - 1

    return result


#
#
#


@_intrinsic
def max_int(arg: int | Unsigned):
    if isinstance(arg, int):
        return arg
    else:
        return TypeQualifierBase.decay(arg).max_int()


@_intrinsic
def int_log_2(inp: int) -> int:
    assert isinstance(inp, int), "argument must be an integer"
    assert inp > 0, "argument must be larger than 0"
    assert inp.bit_count() == 1, "argument must be a power of 2"
    return inp.bit_length() - 1


@_intrinsic
def ceil_log_2(inp: int) -> int:
    assert isinstance(inp, int), "argument must be an integer"
    assert inp > 0, "argument must be larger than 0"

    if inp.bit_count() == 1:
        return inp.bit_length() - 1
    return inp.bit_length()


@_intrinsic
def is_pow_two(inp: int):
    return inp.bit_count() == 1


@_intrinsic
def _is_one(val):
    x = isinstance(val, int) and val == 1
    return x


async def tick() -> None:
    await true


async def wait_for(duration: int | Unsigned | Duration, *, allow_zero: bool = False):
    if isinstance(duration, Duration):
        ctx = SequentialContext.current()
        assert (
            ctx is not None
        ), "wait_for can only infer the clock in sequential contexts created with a Clock with defined frequency"
        cnt = duration.count_periods(ctx.clk().period())
    else:
        cnt = duration

    if allow_zero:
        if duration == 0:
            return
    else:
        assert (
            cnt > 0
        ), "waiting for 0 ticks only possible when allow_zero is set to True"

    if _is_one(cnt):
        await true
    else:
        counter = Signal[Unsigned.upto(max_int(cnt - 1))](cnt - 1)
        while counter:
            counter <<= counter - 1


async def wait_forever():
    await false


class Waiter:
    @_intrinsic
    def _as_ticks(self, duration: int | Duration):
        if isinstance(duration, int) or instance_check(duration, Unsigned):
            return duration
        ctx = SequentialContext.current()
        return duration.count_periods(ctx.clk().period())

    @_intrinsic
    def _init(self):
        if self._max_duration_cnt is not None:
            return

        self._max_duration_cnt = self._as_ticks(self._max_duration)
        self._duration_cnt = Signal[Unsigned.upto(self._max_duration_cnt)](Null)

    def __init__(self, max_duration: int | Duration):
        self._max_duration = max_duration
        self._max_duration_cnt: int | None = None
        self._duration_cnt: Signal[Unsigned]

    async def wait_for(
        self, duration: int | Unsigned | Duration, *, allow_zero: bool = False
    ):
        self._init()
        cnt = self._as_ticks(duration)

        assert (
            cnt <= self._max_duration_cnt
        ), "duration exceeds max_duration set in constructor"

        if allow_zero:
            if cnt == 0:
                return
        else:
            assert (
                cnt > 0
            ), "waiting for 0 ticks only possible when allow_zero is set to True"

        if _is_one(cnt):
            await true
        else:
            self._duration_cnt <<= cnt - 1
            while self._duration_cnt:
                self._duration_cnt <<= self._duration_cnt - 1


class OutShiftRegister:
    def __init__(
        self, src: BitVector, msb_first=False, unchecked=False, initial_empty=False
    ):
        self._unchecked = unchecked
        self._msb_first = msb_first

        if unchecked:
            self._data = Signal(src, name=name("out_shift_reg"))
        elif initial_empty:
            self._data = Signal[BitVector[src.width]](name="out_shift_reg")
        else:
            if msb_first:
                self._data = Signal(src @ Bit(True), name=name("out_shift_out"))
            else:
                self._data = Signal(Bit(True) @ src, name=name("out_shift_out"))

    def set_data(self, data):
        if self._unchecked:
            self._data <<= data
        else:
            assert len(data) == len(self._data) - 1
            if self._msb_first:
                self._data <<= Signal(data @ Bit(True))
            else:
                self._data <<= Signal(Bit(True) @ data)

    async def shift_all(self, target: Bit | BitVector, shift_delayed=False):
        static_assert(
            not self._unchecked,
            "the shift_all method cannot be used on unchecked shift registers",
        )
        count = target.width if instance_check(target, BitVector) else None

        if not shift_delayed:
            target <<= self.shift(count)

        while not self.empty():
            target <<= self.shift(count)

    def empty(self):
        static_assert(
            not self._unchecked,
            "the empty method cannot be used on unchecked shift registers",
        )
        if self._msb_first:
            return not self._data.lsb(rest=1)
        else:
            return not self._data.msb(rest=1)

    def shift(self, count: int | None = None):
        shift_width = count if count is not None else 1
        assert (
            isinstance(shift_width, int) and shift_width > 0
        ), "count must be a constant positive integer value"

        if self._msb_first:
            after_shift = self._data.lsb(rest=shift_width) @ zeros(shift_width)

            # check, that the marker bit is never shifted out of the extended register
            assert self._unchecked or bool(
                after_shift
            ), "invalid shift, register already empty"
            self._data <<= after_shift
            return self._data.msb(count)
        else:
            after_shift = zeros(shift_width) @ self._data.msb(rest=shift_width)
            assert self._unchecked or bool(
                after_shift
            ), "invalid shift, register already empty"
            self._data <<= after_shift
            return self._data.lsb(count)


class InShiftRegister:
    def __init__(self, len: int, msb_first=False, unchecked=False):
        self._unchecked = unchecked
        self._msb_first = msb_first
        self._len = len

        if unchecked:
            self._data = Signal[BitVector[len]](Null, name=name("in_shift_reg"))
        else:
            if msb_first:
                self._data = Signal(
                    BitVector[len](Null) @ Bit(True), name=name("in_shift_reg")
                )
            else:
                self._data = Signal(
                    Bit(True) @ BitVector[len](Null), name=name("in_shift_reg")
                )

    async def shift_all(self, src: Bit | BitVector, shift_delayed=False):
        static_assert(
            not self._unchecked,
            "the shift_all method cannot be used on unchecked shift registers",
        )

        if not shift_delayed:
            self.shift(src)

        while not self.full():
            self.shift(src)

        return self.data()

    def clear(self):
        if self._unchecked:
            self._data <<= Null
        else:
            if self._msb_first:
                self._data <<= BitVector[self._len](Null) @ Bit(True)
            else:
                self._data <<= Bit(True) @ BitVector[self._len](Null)

    def full(self):
        static_assert(
            not self._unchecked,
            "the full method cannot be used on unchecked shift registers",
        )

        if self._msb_first:
            return self._data.msb().copy()
        else:
            return self._data.lsb().copy()

    def shift(self, src: Bit | BitVector):
        shift_cnt = width(src)

        if self._msb_first:
            assert self._unchecked or not self._data.msb(
                shift_cnt
            ), "invalid shift, register already full"
            self._data <<= self._data.lsb(rest=shift_cnt) @ src
        else:
            assert self._unchecked or not self._data.lsb(
                shift_cnt
            ), "invalid shift, register already full"
            self._data <<= src @ self._data.msb(rest=shift_cnt)

    def data(self):
        if self._unchecked:
            return self._data.copy()
        else:
            if self._msb_first:
                return self._data.lsb(rest=1).copy()
            else:
                return self._data.msb(rest=1).copy()


def continuous_counter(
    ctx: SequentialContext, limit, *, on_change=nop, start_at_limit=False
):
    if max_int(limit) == 0:

        @ctx
        def proc():
            on_change(Unsigned[1](0))

        return Signal[Unsigned[1]](0)
    else:
        if not start_at_limit:
            counter = Signal[Unsigned.upto(max_int(limit))](0, name=name("counter"))

        if start_at_limit:
            assert not is_qualified(limit)
            counter = Signal[Unsigned.upto(max_int(limit))](limit, name=name("counter"))

            def reset_fn():
                counter.next = limit

            ctx = ctx(on_reset=reset_fn)

        @ctx
        def proc():
            nonlocal counter

            if not is_qualified(limit):
                next_value = 0 if counter == limit else (counter + 1)
            else:
                next_value = 0 if counter >= limit else (counter + 1)

            counter <<= next_value
            on_change(next_value)

        return counter


class ToggleSignal:
    def __init__(
        self,
        ctx: SequentialContext,
        first_duration: int | Unsigned | Duration,
        second_duration: int | Unsigned | Duration | None = None,
        *,
        default_state: bool = False,
        first_state: bool = False,
        require_enable: bool = False,
        on_rising=None,
        on_falling=None,
        _prefix="toggle",
    ):
        with prefix(_prefix):
            assert not is_qualified(
                default_state
            ), "default_state must be runtime constant"

            first = first_duration
            second = second_duration if second_duration is not None else first

            if isinstance(first, Duration):
                cnt_first = first.count_periods(ctx.clk().period())
            else:
                cnt_first = first

            if isinstance(second, Duration):
                cnt_second = second.count_periods(ctx.clk().period())
            else:
                cnt_second = second

            if is_qualified(cnt_first) or is_qualified(cnt_second):
                max_counter_end = max_int(cnt_first) + max_int(cnt_second) - 1
                CounterType = Unsigned.upto(max_counter_end)
                counter_end = Signal[CounterType](name="counter_end")

                @concurrent_context
                def logic():
                    sum = Value[CounterType](cnt_first) + Value[CounterType](cnt_second)
                    assert sum != 0, "counter end was set to 0"

                    # cast cnt_first and cnt_second to CounterType
                    # to avoid unsigned overflow
                    counter_end.next = sum - 1

            else:
                counter_end = cnt_first + cnt_second - 1
                assert counter_end >= 0, "toggle period may not be zero"

            self._reset_counter = Signal[Bit](require_enable, name=name("reset"))

            self._state = Signal[Bit](default_state, name=name("state"))
            self._rising = Signal[Bit](False, name=name("rising"))
            self._falling = Signal[Bit](False, name=name("falling"))

            def change_handler(next_cnt):
                if first_state:
                    next_state = next_cnt < cnt_first
                else:
                    next_state = not (next_cnt < cnt_first)

                self._state <<= next_state

                rising = not self._state and next_state
                falling = self._state and not next_state

                self._rising <<= rising
                self._falling <<= falling

                if on_rising is not None and rising:
                    on_rising()
                if on_falling is not None and falling:
                    on_falling()

            continuous_counter(
                ctx.or_reset(self._reset_counter), counter_end, on_change=change_handler
            )

    def get_reset_signal(self):
        return self._reset_counter

    def enable(self):
        self._reset_counter <<= False

    def disable(self):
        self._reset_counter <<= True

    def rising(self):
        return self._rising

    def falling(self):
        return self._falling

    def state(self):
        return self._state


class ClockDivider:
    def __init__(
        self,
        ctx: SequentialContext,
        duration: int | Unsigned | Duration,
        *,
        default_state: bool = False,
        tick_at_start: bool = False,
        require_enable: bool = False,
        on_rising=None,
        on_falling=None,
        _prefix="clkdiv",
    ):
        with prefix(_prefix):
            assert not is_qualified(
                default_state
            ), "default_state must be runtime constant"

            if isinstance(duration, Duration):
                cnt_duration = duration.count_periods(ctx.clk().period())
            else:
                cnt_duration = duration

            if is_qualified(cnt_duration):
                max_counter_end = max_int(cnt_duration) - 1
                CounterType = Unsigned.upto(max_counter_end)
                counter_end = Signal[CounterType](name="counter_end")

                @concurrent_context
                def logic():
                    # cast cnt_first and cnt_second to CounterType
                    # to avoid unsigned overflow
                    counter_end.next = cnt_duration - 1
                    assert cnt_duration >= 1, "clkdiv period must be greater than 1"

            else:
                counter_end = cnt_duration - 1
                assert counter_end >= 1, "clockdiv period must be greater than 1"

            self._reset_counter = Signal[Bit](require_enable, name=name("reset"))

            self._ctx = ctx
            self._state = Signal[Bit](default_state, name=name("state"))
            self._rising = Signal[Bit](False, name=name("rising"))
            self._falling = Signal[Bit](False, name=name("falling"))

            def change_handler(next_cnt):
                next_state = (not default_state) if next_cnt == 0 else default_state

                self._state <<= next_state

                rising = not self._state and next_state
                falling = self._state and not next_state

                self._rising <<= rising
                self._falling <<= falling

                if on_rising is not None and rising:
                    on_rising()
                if on_falling is not None and falling:
                    on_falling()

            continuous_counter(
                ctx.or_reset(self._reset_counter),
                counter_end,
                on_change=change_handler,
                start_at_limit=tick_at_start,
            )

    def get_reset_signal(self):
        return self._reset_counter

    def enable(self):
        self._reset_counter <<= False

    def disable(self):
        self._reset_counter <<= True

    def rising(self):
        return self._rising

    def falling(self):
        return self._falling

    def state(self):
        return self._state


_prefix_name = name


class SyncFlag:
    @_intrinsic
    def _cmp_rx(self):
        if self._rx_delay == 0:
            return self._rx

        ctx = SequentialContext.current()

        if ctx is None or ctx is self._tx_ctx:
            return self._rx

        if ctx is self._rx_ctx:
            return self._set_rx

        if ctx is self._rx_indirect_owner:
            return self._rx_indirect

        at_end_of_context(self._impl_rx_indirect)
        self._rx_indirect = Signal[Bit](name=self._rx_indirect_name)
        self._rx_indirect_owner = ctx
        return self._rx_indirect

    @_intrinsic
    def _cmp_tx(self):
        if self._tx_delay == 0:
            return self._tx

        ctx = SequentialContext.current()

        if ctx is None or ctx is self._rx_ctx:
            return self._tx

        if ctx is self._tx_ctx:
            return self._set_tx

        if ctx is self._tx_indirect_owner:
            return self._tx_indirect

        at_end_of_context(self._impl_tx_indirect)
        self._tx_indirect = Signal[Bit](name=self._tx_indirect_name)
        self._tx_indirect_owner = ctx
        return self._tx_indirect

    async def _impl_rx_indirect(self):
        ctx = SequentialContext.current()

        with always:
            if ctx is self._rx_ctx:
                self._rx_indirect <<= self._set_rx
            else:
                self._rx_indirect <<= self._rx

    async def _impl_tx_indirect(self):
        ctx = SequentialContext.current()

        with always:
            if ctx is self._tx_ctx:
                self._tx_indirect <<= self._set_tx
            else:
                self._tx_indirect <<= self._tx

    async def _impl_tx_delayline(self):
        # delay-1 because assignment to _tx adds one
        self._tx <<= delayed(self._set_tx, self._tx_delay - 1, initial=Null)

    async def _impl_rx_delayline(self):
        # delay-1 because assignment to _rx adds one
        self._rx <<= delayed(self._set_rx, self._rx_delay - 1, initial=Null)

    def _impl_rx_delay(self):
        # rx delay is implemented in tx_ctx

        tx_ctx = SequentialContext.current()

        if self._tx_ctx is None:
            as_pyeval(setattr, self, "_tx_ctx", tx_ctx)

            if self._rx_delay != 0:
                assert (
                    self._rx_ctx is not tx_ctx
                ), "std.SyncFlag with delay cannot be set and cleared in the same context"
                at_end_of_context(self._impl_rx_delayline)

            # return True first time context is detected so
            # higher level abstractions can define additional logic
            return True
        else:
            assert tx_ctx is self._tx_ctx, "std.SyncFlag set in more than one context"
            return False

    def _impl_tx_delay(
        self,
    ):
        # tx delay is implemented in rx_ctx

        rx_ctx = SequentialContext.current()

        if self._rx_ctx is None:
            as_pyeval(setattr, self, "_rx_ctx", rx_ctx)

            if self._tx_delay != 0:
                assert (
                    self._tx_ctx is not rx_ctx
                ), "std.SyncFlag with delay cannot be set and cleared in the same context"
                at_end_of_context(self._impl_tx_delayline)

            # return True first time context is detected so
            # higher level abstractions can define additional logic
            return True
        else:
            assert rx_ctx is self._rx_ctx, "std.SyncFlag set in more than one context"
            return False

    def __init__(self, *, name="sync_flag", delay=None, tx_delay=None, rx_delay=None):
        if delay is not None:
            assert (
                tx_delay is rx_delay is None
            ), "tx_delay/rx_delay cannot be set when delay is specified"
            self._tx_delay = delay
            self._rx_delay = delay
        else:
            self._tx_delay = tx_delay if tx_delay is not None else 0
            self._rx_delay = rx_delay if rx_delay is not None else 0

        with prefix(name) as p:
            self._tx = Signal[Bit](False, name=p.name("tx"))

            if self._tx_delay == 0:
                self._set_tx = self._tx
            else:
                self._set_tx = Signal[Bit](False, name=p.name("set_tx"))

            self._rx = Signal[Bit](False, name=p.name("rx"))

            if self._rx_delay == 0:
                self._set_rx = self._rx
            else:
                self._set_rx = Signal[Bit](False, name=p.name("set_rx"))

            self._tx_ctx = None
            self._rx_ctx = None

            if self._rx_delay != 0 or self._tx_delay != 0:
                self._tx_indirect_name = p.name("tx_indirect")
                self._rx_indirect_name = p.name("rx_indirect")

                self._rx_indirect_owner = None
                self._tx_indirect_owner = None
                self._rx_indirect = None
                self._tx_indirect = None

    def set(self):
        self._impl_rx_delay()
        self._set_tx <<= ~self._rx

    def clear(self):
        self._impl_tx_delay()
        self._set_rx <<= self._tx

    @expr_fn
    def is_set(self):
        return self._cmp_tx() != self._cmp_rx()

    @expr_fn
    def is_clear(self):
        return self._cmp_tx() == self._cmp_rx()

    async def receive(self):
        await self.is_set()
        self.clear()

    async def __aenter__(self):
        await self.is_set()

    async def __aexit__(self, val, type, traceback):
        self.clear()


_MailboxType = template_arg.Type


class Mailbox(Template[_MailboxType]):
    _datatype: _MailboxType

    def __init__(self, *, delay=None, tx_delay=None, rx_delay=None):
        with prefix("mailbox"):
            self._data = NamedQualifier[Signal, "data"][self._datatype]()
            self._flag = SyncFlag(delay=delay, tx_delay=tx_delay, rx_delay=rx_delay)

    def send(self, data):
        self._data <<= data
        self._flag.set()

    async def receive(self):
        async with self._flag:
            return self._data

    def data(self):
        return self._data

    @expr_fn
    def is_set(self):
        return self._flag.is_set()

    @expr_fn
    def is_clear(self):
        return self._flag.is_clear()

    def clear(self):
        return self._flag.clear()


#
#
#


class _FifoArgs:
    def __init__(self, arg):
        self.elemtype, self.count = arg
        assert isinstance(self.elemtype, type)
        assert isinstance(self.count, int)

    def __hash__(self) -> int:
        return hash((self.elemtype, self.count))

    def __eq__(self, other: _FifoArgs) -> bool:
        return self.elemtype is other.elemtype and self.count == other.count


class Fifo(Template[_FifoArgs]):
    _elemtype_: _FifoArgs.elemtype
    _count_: _FifoArgs.count

    async def _impl_sync_read_index(self):
        if self._sync_flag.is_clear():
            self._buf_write_index <<= self._set_write_index
            self._read_index <<= self._buf_read_index
            self._sync_flag.set()

    async def _impl_sync_write_index(self):
        if self._sync_flag.is_set():
            self._buf_read_index <<= self._set_read_index
            self._write_index <<= self._buf_write_index
            self._sync_flag.clear()

    @_intrinsic
    def _cmp_full(self):
        if not self._sync_contexts:
            return self._full

        ctx = SequentialContext.current()

        if ctx is None:
            return self._full

        if ctx is self._sync_flag._tx_ctx:
            return self._full_in_sender

        if ctx is self._sync_flag._rx_ctx:
            return self._full_in_receiver

        if ctx is self._full_indirect_owner:
            return self._full_indirect

        at_end_of_context(self._impl_full_indirect)
        self._full_indirect = Signal[Bit](name=self._full_indirect_name)
        return self._full_indirect

    @_intrinsic
    def _cmp_empty(self):
        if not self._sync_contexts:
            return self._empty

        ctx = SequentialContext.current()

        if ctx is None:
            return self._empty

        if ctx is self._sync_flag._tx_ctx:
            return self._empty_in_sender

        if ctx is self._sync_flag._rx_ctx:
            return self._empty_in_receiver

        if ctx is self._empty_indirect_owner:
            return self._empty_indirect

        at_end_of_context(self._impl_empty_indirect)
        self._empty_indirect = Signal[Bit](name=self._empty_indirect_name)
        return self._empty_indirect

    async def _impl_full_indirect(self):
        with always:
            ctx = SequentialContext.current()

            if ctx is self._sync_flag._tx_ctx:
                self._full_indirect <<= self._full_in_sender
            elif ctx is self._sync_flag._rx_ctx:
                self._full_indirect <<= self._full_in_receiver
            else:
                self._full_indirect <<= self._full

    async def _impl_empty_indirect(self):
        with always:
            ctx = SequentialContext.current()

            if ctx is self._sync_flag._tx_ctx:
                self._empty_indirect <<= self._empty_in_sender
            elif ctx is self._sync_flag._rx_ctx:
                self._empty_indirect <<= self._empty_in_receiver
            else:
                self._empty_indirect <<= self._empty

    #
    #
    #

    def _next_index(self, index):
        if is_pow_two(self._max_index + 1):
            return index + 1
        else:
            return index + 1 if index != self._max_index else 0

    @_intrinsic
    def __len__(self):
        return self._count_

    def __init__(self, name="fifo", delay=None, rx_delay=None, tx_delay=None):
        count = self._count_
        self._max_index = count - 1
        CounterType = Unsigned.upto(self._max_index)

        if delay is not None:
            assert (
                tx_delay is rx_delay is None
            ), "tx_delay/rx_delay cannot be set when delay is specified"
            self._tx_delay = delay
            self._rx_delay = delay
        else:
            self._tx_delay = tx_delay if tx_delay is not None else 0
            self._rx_delay = rx_delay if rx_delay is not None else 0

        self._sync_contexts = self._rx_delay != 0 or self._tx_delay != 0

        with prefix(name) as p:
            self._mem = Signal[Array[self._elemtype_, count]](name=p.name("mem"))

            self._write_index = Signal[CounterType](0, name=p.name("wr_index"))
            self._read_index = Signal[CounterType](0, name=p.name("rd_index"))

            self._empty = Signal[Bit](name=p.name("empty"))
            self._full = Signal[Bit](name=p.name("full"))

            if self._sync_contexts:
                self._sync_flag = SyncFlag(
                    name="sync", rx_delay=self._rx_delay, tx_delay=self._tx_delay
                )

                self._set_write_index = Signal[CounterType](
                    0, name=p.name("set_wr_index")
                )

                self._set_read_index = Signal[CounterType](
                    0, name=p.name("set_rd_index")
                )

                self._buf_write_index = Signal[CounterType](
                    0, name=p.name("buf_wr_index")
                )

                self._buf_read_index = Signal[CounterType](
                    0, name=p.name("buf_rd_index")
                )

                self._full_in_sender = Signal[Bit](name=p.name("full_in_sender"))
                self._empty_in_sender = Signal[Bit](name=p.name("empty_in_sender"))
                self._full_in_receiver = Signal[Bit](name=p.name("full_in_receiver"))
                self._empty_in_receiver = Signal[Bit](name=p.name("empty_in_receiver"))

                self._full_indirect_name = p.name("full_indirect")
                self._empty_indirect_name = p.name("empty_indirect")

                self._full_indirect_owner = None
                self._empty_indirect_owner = None

                @concurrent
                def logic():
                    self._full_in_sender <<= (
                        self._next_index(self._set_write_index) == self._read_index
                    )

                    self._empty_in_sender <<= self._set_write_index == self._read_index

                    self._full_in_receiver <<= (
                        self._next_index(self._write_index) == self._set_read_index
                    )

                    self._empty_in_receiver <<= (
                        self._write_index == self._set_read_index
                    )

            else:
                a = a = 0
                self._set_write_index = self._write_index
                self._set_read_index = self._read_index

        @concurrent
        def logic():
            self._empty <<= self._write_index == self._read_index
            self._full <<= self._next_index(self._write_index) == self._read_index

    def push(self, data):
        if self._sync_contexts:
            if self._sync_flag._impl_rx_delay():
                at_end_of_context(self._impl_sync_read_index)

        assert not self._full, "writing to full fifo"
        self._mem[self._set_write_index] <<= data
        self._set_write_index <<= self._next_index(self._set_write_index)

    def pop(self):
        if self._sync_contexts:
            if self._sync_flag._impl_tx_delay():
                at_end_of_context(self._impl_sync_write_index)

        assert not self._empty, "reading from empty fifo"
        self._set_read_index <<= self._next_index(self._set_read_index)
        return self._mem[self._set_read_index]

    def front(self):
        return self._mem[self._set_read_index]

    def empty(self):
        return self._cmp_empty()

    def full(self):
        return self._cmp_full()



import unittest
import random

from cohdl import BitVector, BitOrder, Bit, Null, Full
from cohdl._core._type_qualifier import (
    TypeQualifier,
    Variable,
    Signal,
    Temporary,
)


def rnd_bit_string(length):
    return "".join(["01U"[random.randint(0, 2)] for _ in range(length)])


class BitVectorTester(unittest.TestCase):
    def test_construction(self):
        def test(width, default=None, expected_assert=None):
            if expected_assert is not None:
                self.assertRaises(
                    expected_assert, lambda d: BitVector[width](d), default
                )
            else:
                a = BitVector[width](default)
                self.assertTrue(a.order is BitOrder.DOWNTO)
                self.assertTrue(a.width == width)
                self.assertTrue(len(a) == width)

                if isinstance(default, str):
                    self.assertTrue(default == str(a))

                b = BitVector[width](a)

                self.assertTrue(str(a) == str(b))

                for i in range(width):
                    self.assertTrue(isinstance(a[i], Bit))
                    self.assertTrue(isinstance(b[i], Bit))

                    self.assertTrue(a[i] == b[i])
                    self.assertFalse(a[i] is b[i])

                self.assertIsNot(a, b)

                rnd = BitVector[width](rnd_bit_string(width))

                prev_bits = [b[i] for i in range(width)]

                b._assign(rnd)

                for i in range(width):
                    # check, that subscript access returns reference
                    self.assertTrue(b[i] is prev_bits[i])
                    self.assertTrue(b[i] == rnd[i])
                    self.assertTrue(b[i] is not rnd[i])

                self.assertTrue(b == rnd)
                self.assertTrue(not b != rnd)

                b._assign(rnd)

                for i in range(width):
                    self.assertTrue(b[i] is prev_bits[i])

                self.assertTrue(b == rnd)
                self.assertTrue(rnd == b)

                for i in range(width):
                    self.assertTrue(b[i] is prev_bits[i])

                a._assign(rnd)

                self.assertTrue(b == a)
                self.assertTrue(a == b)

                self.assertRaises(AssertionError, b._assign, BitVector[width + 1]())

        for w in [1, 2, 3, 4, 5, 6, 7, 8, 16, 32, 37, 128, 1024, 0, -1, -2, -5, -100]:
            if w < 0:
                test(w, expected_assert=AssertionError)
                test(w, rnd_bit_string(w), expected_assert=AssertionError)
            else:
                test(w)
                test(w, rnd_bit_string(w))

                if w != 0:
                    test(w, " " * w, expected_assert=AssertionError)

                zeros = BitVector[w](Null)
                ones = BitVector[w](Full)
                rnd = BitVector[w](rnd_bit_string(w))

                test(w, zeros)
                test(w, ones)
                test(w, rnd)

    def test_cmp(self):
        for width in [1, 2, 4, 10, 16, 471, 1024]:
            ones = BitVector[width](Full)
            zeros = BitVector[width](Null)

            self.assertFalse(ones == zeros)
            self.assertFalse(zeros == ones)
            self.assertTrue(ones != zeros)
            self.assertTrue(zeros != ones)

    def test_msb_lsb_single(self):
        for category in [Signal, Variable, Temporary]:
            for order in [BitOrder.DOWNTO, BitOrder.UPTO]:
                for width in [1, 2, 4, 16, 423, 1024]:
                    v = category[BitVector[width]]()

                    lsb = v.lsb()
                    msb = v.msb()
                    left = v.left()
                    right = v.right()

                    self.assertTrue(issubclass(category, TypeQualifier))

                    self.assertIs(type(lsb).qualifier, category)
                    self.assertIs(type(msb).qualifier, category)
                    self.assertIs(type(left).qualifier, category)
                    self.assertIs(type(right).qualifier, category)

                    self.assertIs(type(lsb), category[Bit])
                    self.assertIs(type(msb), category[Bit])
                    self.assertIs(type(left), category[Bit])
                    self.assertIs(type(right), category[Bit])

                    self.assertIs(lsb._root, v)

                    lsb = lsb.get()
                    msb = msb.get()
                    left = left.get()
                    right = right.get()

                    self.assertTrue(isinstance(lsb, Bit))
                    self.assertTrue(isinstance(msb, Bit))
                    self.assertTrue(isinstance(left, Bit))
                    self.assertTrue(isinstance(right, Bit))

                    if width == 1:
                        self.assertTrue(lsb is msb is left is right)
                    else:
                        self.assertTrue(lsb is not msb)
                        self.assertTrue(left is not right)

                    if order is BitOrder.DOWNTO:
                        self.assertTrue(lsb is right)
                        self.assertTrue(msb is left)
                        self.assertTrue(lsb is v[0].get())
                        self.assertTrue(msb is v[width - 1].get())
                    else:
                        pass

    def test_msb_lsb_slice(self):
        v = BitVector[16]()

        self.assertRaises(AssertionError, v.lsb, 0)
        self.assertRaises(AssertionError, v.msb, 0)
        self.assertRaises(AssertionError, v.left, 0)
        self.assertRaises(AssertionError, v.right, 0)

        for i in range(1, v.width + 1):
            lsb = v.lsb(i)
            msb = v.msb(i)
            left = v.left(i)
            right = v.right(i)

            self.assertTrue(isinstance(lsb, BitVector))
            self.assertTrue(isinstance(msb, BitVector))
            self.assertTrue(isinstance(left, BitVector))
            self.assertTrue(isinstance(right, BitVector))

            self.assertTrue(lsb.order is BitOrder.DOWNTO)

            alias_lsb = right
            alias_msb = left

            for a, b in zip(lsb._value, alias_lsb._value):
                self.assertTrue(a is b)

            for a, b in zip(msb._value, alias_msb._value):
                self.assertTrue(a is b)

    def test_concat(self):
        for a in [1, 2, 5, 16]:
            for b in [1, 2, 5, 16]:
                str_a = rnd_bit_string(a)
                str_b = rnd_bit_string(b)

                bv_a = BitVector[a](str_a)
                bv_b = BitVector[b](str_b)

                concat = bv_a @ bv_b

                self.assertTrue(concat.width == a + b)
                self.assertTrue(concat == BitVector[a + b](str_a + str_b))

                bv_a._assign(BitVector[a](Null))
                bv_b._assign(BitVector[b](Null))

                self.assertTrue(
                    concat == BitVector[a + b](str_a + str_b),
                    "changing source does not affect output of operation",
                )

                concat._assign(BitVector[a + b](rnd_bit_string(a + b)))

                self.assertTrue(
                    bv_a == BitVector[a](Null),
                    "changing result does not affect input of operation",
                )
                self.assertTrue(
                    bv_b == BitVector[b](Null),
                    "changing result does not affect input of operation",
                )

        for a in [1, 2, 5, 16]:
            for bit in [Bit(0), Bit(1)]:
                rnd_str = rnd_bit_string(a)
                bv = BitVector[a](rnd_str)
                bit_left = bit @ bv
                bit_right = bv @ bit

                self.assertTrue(bit_left.width == a + 1)
                self.assertTrue(bit_right.width == a + 1)

                if bit:
                    self.assertTrue(bit_left == BitVector[a + 1]("1" + rnd_str))
                    self.assertTrue(bit_right == BitVector[a + 1](rnd_str + "1"))
                else:
                    self.assertTrue(bit_left == BitVector[a + 1]("0" + rnd_str))
                    self.assertTrue(bit_right == BitVector[a + 1](rnd_str + "0"))



from __future__ import annotations
from _ast import Lambda
from cohdl.utility.source_location import SourceLocation
from cohdl.utility.virtual_traceback import VirtualFrame


from abc import abstractmethod

import ast
import inspect
import textwrap
from typing import Any, Iterable, Tuple


class _Unbound:
    """
    singleton that indicates a local variables name that is not bound to a specific value
    """


def optional(arg, generator):
    return generator() if arg is None else arg


class _ClassifyNames(ast.NodeVisitor):
    """
    process a block of code and determine, which names are local
    and which are inherited from the parent scope
    """

    def visit_Name(self, node: ast.Name):
        if isinstance(node.ctx, ast.Load):
            self.used_names.add(node.id)
        elif isinstance(node.ctx, ast.Store):
            self.used_names.add(node.id)

            if node.id not in self._explicit_nonlocal:
                self.local_names.add(node.id)
        else:
            raise AssertionError(f"invalid context for name `{node.id}`")

    def visit_Nonlocal(self, node: ast.Nonlocal) -> Any:
        for name in node.names:
            assert (
                name not in self.used_names
            ), f"name {name} used before nonlocal declaration"
            self.used_names.add(name)
            self._explicit_nonlocal.add(name)

    def visit_ClassDef(self, node: ast.ClassDef):
        raise AssertionError(f"class definition in local function not supported")

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
        self.visit_FunctionDef(node)

    def _visit_fn_or_lambda(self, node: ast.FunctionDef | ast.Lambda):
        args = node.args

        self.generic_visit_list(args.defaults)
        self.generic_visit_list(args.kw_defaults)

        subargs = [arg.arg for arg in [*args.posonlyargs, *args.args, *args.kwonlyargs]]

        if args.kwarg is not None:
            subargs.append(args.kwarg.arg)
        if args.vararg is not None:
            subargs.append(args.vararg.arg)

        sub_visitor = _ClassifyNames(subargs, node.body)

        self.used_names.update(sub_visitor.nonlocals())

    def visit_FunctionDef(self, node: ast.FunctionDef | ast.AsyncFunctionDef):
        self.used_names.add(node.name)
        self.local_names.add(node.name)

        self._visit_fn_or_lambda(node)

        self.generic_visit_list(node.decorator_list)

    def visit_Lambda(self, node: Lambda) -> Any:
        self._visit_fn_or_lambda(node)

    def __init__(self, args, body: list | ast.AST):
        self.used_names: set[str] = set()
        self.local_names: set[str] = set(args)
        self._explicit_nonlocal: set[str] = set()

        # explicitly mark __class__ as used
        # (required by super() call)
        self.used_names.add("__class__")
        super().__init__()

        if isinstance(body, list):
            self.generic_visit_list(body)
        else:
            self.visit(body)

    def generic_visit_list(self, inp: list):
        for stmt in inp:
            self.visit(stmt)

    def all_names(self):
        return self.used_names | self.local_names

    def local(self):
        return self.local_names

    def nonlocals(self):
        return self.used_names - self.local_names


class _ScopeBase:
    def _capture_env(self, local_names, nonlocal_names, global_dict, nonlocal_dict):
        result = {}

        builtins = global_dict["__builtins__"] if "__builtins__" in global_dict else {}

        for name in local_names:
            result[name] = _Unbound

        for name in nonlocal_names:
            if name in nonlocal_dict:
                result[name] = nonlocal_dict[name]
            elif name in global_dict:
                result[name] = global_dict[name]
            elif isinstance(builtins, dict) and name in builtins:
                result[name] = builtins[name]
            elif hasattr(builtins, name):
                result[name] = getattr(builtins, name)
            else:
                # special case for possibly unbound cell __class__
                assert name == "__class__", f"invalid name '{name}'"

        return result

    @abstractmethod
    def capture(self) -> dict[str, Any]:
        """
        returns a dictionary that contains all names used in the scope,
        local names are set to _Unbound, function arguments are treated like globals
        """
        ...


class ScopeRef(_ScopeBase):
    def __init__(
        self,
        local_names: Iterable[str],
        nonlocal_names: Iterable[str],
        global_dict: dict,
        nonlocal_dict: dict,
    ):
        self._scope = self._capture_env(
            local_names, nonlocal_names, global_dict, nonlocal_dict
        )

    def capture(self) -> dict[str, Any]:
        return self._scope


class FunctionDefinition:
    # A mapping over all previously parsed function definitions.
    # The id() of function objects is used as a key.
    # The values are tuples of FunctionDefinition and the function
    # object to prevent possible key collisions after garbage collection
    # of the function.
    _known_definitions: dict[int, tuple[FunctionDefinition, Any]] = {}

    @staticmethod
    def from_ast_body(
        fn: list[ast.stmt],
        name: str,
        scope_ref: _ScopeBase,
        is_async: bool,
        posonly: list[str] | None = None,
        args: list[str] | None = None,
        kwonly: list[str] | None = None,
        defaults: dict | None = None,
        kwdefaults: dict | None = None,
        vararg: str | None = None,
        kwarg: str | None = None,
        self_arg=_Unbound,
        *,
        location: SourceLocation,
    ):
        posonly = optional(posonly, list)
        args = optional(args, list)
        kwonly = optional(kwonly, list)
        defaults = optional(defaults, dict)
        kwdefaults = optional(kwdefaults, dict)

        return FunctionDefinition(
            fn,
            name,
            scope_ref,
            is_async,
            posonly=posonly,
            args=args,
            kwonly=kwonly,
            defaults=defaults,
            kwdefaults=kwdefaults,
            vararg=vararg,
            kwarg=kwarg,
            self_arg=self_arg,
            location=location,
        )

    @staticmethod
    def from_ast_fn(
        fn_def: ast.FunctionDef | ast.AsyncFunctionDef,
        name: str,
        global_dict: dict,
        nonlocal_dict: dict,
        self_arg=_Unbound,
        default_converter=lambda x: x,
        captured_defaults: (
            dict | None
        ) = None,  # if set to None ast representation is used for defaults
        *,
        location: SourceLocation,
    ):
        if isinstance(fn_def, (ast.FunctionDef, ast.Lambda)):
            is_async = False
        elif isinstance(fn_def, ast.AsyncFunctionDef):
            is_async = True
        else:
            raise AssertionError(f"expected function definition not `{fn_def}`")

        posonly: list[str] = [arg.arg for arg in fn_def.args.posonlyargs]
        kwonly: list[str] = [arg.arg for arg in fn_def.args.kwonlyargs]
        args: list[str] = [arg.arg for arg in fn_def.args.args]

        vararg = None if fn_def.args.vararg is None else fn_def.args.vararg.arg
        kwarg = None if fn_def.args.kwarg is None else fn_def.args.kwarg.arg

        fn_args = [*posonly, *kwonly, *args]

        if vararg is not None:
            fn_args.append(vararg)
        if kwarg is not None:
            fn_args.append(kwarg)

        names = _ClassifyNames(fn_args, fn_def.body)

        try:
            scope_ref = ScopeRef(
                names.local_names, names.nonlocals(), global_dict, nonlocal_dict
            )
        except Exception as err:
            frame = VirtualFrame(location, {**global_dict, **nonlocal_dict}, None)
            frame.apply_to_exception(err)
            raise

        if captured_defaults is None:
            defaults = fn_def.args.defaults
            arg_names = [*posonly, *args][-len(defaults) :]
            defaults = {
                name: default_converter(default)
                for name, default in zip(arg_names, defaults)
            }

            kwdefaults = fn_def.args.kw_defaults
            arg_names = kwonly[-len(kwdefaults) :]
            kwdefaults = {
                name: default_converter(default)
                for name, default in zip(arg_names, kwdefaults)
                if default is not None
            }
        else:
            # TODO: check, whether it is ok to use captured_defaults directly without
            # parsing stage (will share Signals/Variables between uses of same function)
            defaults = fn_def.args.defaults
            arg_names = [*posonly, *args][-len(defaults) :] if len(defaults) > 0 else []

            defaults = {name: captured_defaults[name] for name in arg_names}

            kwdefaults = fn_def.args.kw_defaults
            arg_names = kwonly[-len(kwdefaults) :] if len(kwdefaults) > 0 else []

            kwdefaults = {
                name: captured_defaults[name]
                for name, default in zip(arg_names, kwdefaults)
                if default is not None
            }

        if isinstance(fn_def, ast.Lambda):
            # wrap lambda expression in return to make it consistent
            # with normal functions
            body = [ast.Return(fn_def.body)]
        else:
            body = fn_def.body

        return FunctionDefinition.from_ast_body(
            body,
            name,
            scope_ref,
            is_async,
            posonly,
            args,
            kwonly,
            defaults,
            kwdefaults,
            vararg,
            kwarg,
            self_arg=self_arg,
            location=location,
        )

    @staticmethod
    def from_coroutine(coroutine):
        assert inspect.iscoroutine(coroutine)

        if id(coroutine) in FunctionDefinition._known_definitions:
            return FunctionDefinition._known_definitions[id(coroutine)][0]

        def parse_source(source):
            parsed = ast.parse(textwrap.dedent(source))
            assert isinstance(parsed, ast.Module)
            assert len(parsed.body) == 1
            return parsed

        self_arg = _Unbound

        assert not coroutine.cr_running, "running coroutine not allowed in this context"

        code = coroutine.cr_code
        frame = coroutine.cr_frame

        name = coroutine.__name__
        body = parse_source(inspect.getsource(frame)).body[0]
        location = SourceLocation(code.co_filename, code.co_firstlineno, name)

        global_dict = frame.f_globals
        locals_dict = frame.f_locals

        result = FunctionDefinition.from_ast_fn(
            body,
            name,
            global_dict=global_dict,
            nonlocal_dict=locals_dict,
            self_arg=self_arg,
            default_converter=lambda x: x,
            captured_defaults=locals_dict,
            location=location,
        )

        args = []
        kwargs = {}

        for posonly_name in result._posonly:
            args.append(locals_dict[posonly_name])

        for arg_name in result._args:
            args.append(locals_dict[arg_name])

        if result._vararg is not None:
            args.extend(locals_dict[result._vararg])

        for kwarg_name in result._kwonly:
            kwargs[kwarg_name] = locals_dict[kwarg_name]

        for kwonly_name in result._kwonly:
            kwargs[kwarg_name] = locals_dict[kwonly_name]

        if result._kwarg is not None:
            kwargs.update(locals_dict[result._kwarg])

        result = result.bind_args(args, kwargs)

        FunctionDefinition._known_definitions[id(coroutine)] = (result, coroutine)

        return result

    @staticmethod
    def from_callable(callable, default_converter=lambda x: x):
        assert not inspect.iscoroutine(callable)

        if id(callable) in FunctionDefinition._known_definitions:
            return FunctionDefinition._known_definitions[id(callable)][0]

        def parse_source(source):
            parsed = ast.parse(textwrap.dedent(source))
            assert isinstance(parsed, ast.Module)
            assert len(parsed.body) == 1
            return parsed

        self_arg = _Unbound

        if inspect.ismethod(callable):
            self_arg = callable.__self__
            fn = callable.__func__
            name: str = callable.__name__
        elif not inspect.isfunction(callable):
            self_arg = callable
            fn = type(callable).__call__
            name: str = type(callable).__name__
        else:
            fn = callable
            name: str = fn.__name__

        assert inspect.isfunction(fn)

        parsed = parse_source(inspect.getsource(fn))

        if fn.__name__ == "<lambda>":
            lambda_nodes: list[ast.Lambda] = []

            class _SearchNode(ast.NodeVisitor):
                def visit_Lambda(self, node: ast.Lambda) -> Any:
                    lambda_nodes.append(node)
                    return self.generic_visit(node)

            _SearchNode().visit(parsed)

            assert (
                len(lambda_nodes) == 1
            ), "multiple lambdas in same source line not supported"

            body = lambda_nodes[0]
        else:
            body = parsed.body[0]

        assert isinstance(body, (ast.FunctionDef, ast.Lambda))

        closure_vars = inspect.getclosurevars(fn)

        global_dict = fn.__globals__

        signature = inspect.signature(fn)

        captured_defaults = {
            name: value.default
            for name, value in signature.parameters.items()
            if value.default is not value.empty
        }

        code = fn.__code__
        location = SourceLocation(code.co_filename, code.co_firstlineno, name)

        result = FunctionDefinition.from_ast_fn(
            body,
            name,
            global_dict=global_dict,
            nonlocal_dict=closure_vars.nonlocals,
            self_arg=self_arg,
            default_converter=default_converter,
            captured_defaults=captured_defaults,
            location=location,
        )

        FunctionDefinition._known_definitions[id(callable)] = [result, callable]

        return result

    def __init__(
        self,
        body: list[ast.stmt],
        name: str,
        scope_ref: _ScopeBase,
        is_async: bool,
        posonly: list[str],
        args: list[str],
        kwonly: list[str],
        defaults: dict[str, Any],
        kwdefaults: dict[str, Any],
        vararg: str | None,
        kwarg: str | None,
        self_arg=_Unbound,
        *,
        location: SourceLocation,
    ):
        self._location = location

        self._body = body
        self._name = name
        self._scope_ref = scope_ref
        self._is_async = is_async

        self._posonly = posonly
        self._args = args
        self._kwonly = kwonly

        self._vararg = vararg
        self._kwarg = kwarg

        self._defaults = defaults
        self._kwdefaults = kwdefaults

        self._self_arg = self_arg

    def location(self) -> SourceLocation:
        return self._location

    def name(self) -> str:
        return self._name

    def is_async(self):
        return self._is_async

    def body(self):
        return self._body

    def bind_args(self, args: list | Tuple, kwargs: dict) -> InstantiatedFunction:
        """
        assign given arguments to local names in the instantiated function
        """

        args = [*args]

        if self._self_arg is not _Unbound:
            args.insert(0, self._self_arg)

        scope = self._scope_ref.capture()

        # reverse order of args so list.pop can be used to process
        # the arguments in order (pop removes and returns last element of list)
        args = args[::-1]

        class _Dummy: ...

        class _NoSuperArg: ...

        super_arg = _Dummy

        def add_arg(name, val, variadic=False):
            nonlocal super_arg

            if variadic:
                # if the first argument is variadic
                # (*arg or **kwarg) it cannot be interpreted
                # as the object argument of super()
                super_arg = _NoSuperArg
            if super_arg is _Dummy:
                super_arg = val
            scope[name] = val

        for posonly in self._posonly:
            assert (
                posonly not in kwargs or self._kwarg is not None
            ), f"position only argument `{posonly}` may not be passed as keyword argument"

            if len(args) != 0:
                add_arg(posonly, args.pop())
            else:
                add_arg(posonly, self._defaults[posonly])

        for arg in self._args:
            if len(args) != 0:
                assert arg not in kwargs, f"got multiple values for argument `{arg}`"
                add_arg(arg, args.pop())
            elif arg in kwargs:
                add_arg(arg, kwargs[arg])
                del kwargs[arg]
            else:
                if arg in self._defaults:
                    add_arg(arg, self._defaults[arg])
                else:
                    raise AssertionError(f"missing paramerter `{arg}`")

        if self._vararg is None:
            assert len(args) == 0, f"to many arguments passed to function `{args}`"
        else:
            # function has a *vararg parameter, assign remaining
            # non-keyword arguments to this value, undo reversing of arguments from
            # start of function
            add_arg(self._vararg, tuple(args[::-1]))

        for kwonly in self._kwonly:
            if kwonly in kwargs:
                add_arg(kwonly, kwargs[kwonly])
                del kwargs[kwonly]
            else:
                add_arg(kwonly, self._kwdefaults[kwonly])

        if self._kwarg is None:
            assert (
                len(kwargs) == 0
            ), f"to many keyword arguments passed to function `{self._name}`"
        else:
            # function has a **kwarg parameter, assign remaining keyword arguments
            # to this value
            add_arg(self._kwarg, {**kwargs})

        if super_arg is _Dummy or super_arg is _NoSuperArg:
            return InstantiatedFunction(self, scope)

        return InstantiatedFunction(self, scope, super_arg=super_arg)


class InstantiatedFunction:
    def __init__(
        self,
        definition: FunctionDefinition,
        scope: dict,
        super_arg=_Unbound,
    ):
        self._definition = definition
        self._location = definition.location()
        self._scope = scope
        self._super_arg = super_arg

    def location(self):
        return self._definition.location()

    def name(self):
        return self._definition.name()

    def scope(self):
        return self._scope

    def is_async(self) -> bool:
        return self._definition.is_async()

    def body(self):
        return self._definition.body()

    def super_arg(self):
        return self._super_arg



from __future__ import annotations

from cohdl import Bit, BitVector, Unsigned, Signal, Signed, Entity, Port
from cohdl import std

from collections import namedtuple


class TemplateArg:
    def __new__(cls, *field_names: str):
        return namedtuple(cls.__name__, *field_names)


def _template_arg_hash(self):
    return hash(tuple([getattr(self, name) for name in self._field_names]))


def _template_arg_eq(self, other):
    return type(self) is type(other) and all(
        getattr(self, name) == getattr(other, name) for name in self._field_names
    )


def template_arg(*field_names):
    arg_type = namedtuple(*field_names)
    arg_type._field_names = field_names

    arg_type.__hash__ = _template_arg_hash
    arg_type.__eq__ = _template_arg_eq


class MyRecord(std.Record):
    a: Bit
    b: BitVector[8]


class MyEntity(Entity):

    inp_a = Port.input(Bit)
    inp_b = Port.input(BitVector[8])

    out_a = Port.output(Bit)
    out_b = Port.output(BitVector[8])

    def architecture(self):
        arr = std.Array[MyRecord, 8]()
        fifo = std.Fifo[MyRecord, 10]()

        @std.concurrent
        def logic():
            inp_rec = std.Ref[MyRecord](self.inp_a, self.inp_b)
            # out_rec = std.Ref[MyRecord](self.out_a, self.out_b)

            # arr[4] <<= inp_rec

            # arr[7].a <<= inp_rec.a

            # out_rec.a <<= inp_rec.a
            # out_rec.b <<= inp_rec.b

            # out_rec <<= arr[6]

            fifo.push(inp_rec)


print(std.VhdlCompiler.to_string(MyEntity))



from __future__ import annotations

from typing import Any, TypeVar, Generic, NoReturn, Self
import enum

from cohdl import Null, Full, Signal, BitVector, Unsigned, Signed
from cohdl import std

from cohdl.std._context import SequentialContext

from ..utility import SyncFlag

W = TypeVar("W")

# offset
OFF = TypeVar("OFF")
# type
T = TypeVar("T")
# default
D = TypeVar("D")

class GenericArg:
    offset: int
    end: int
    array_step: int

class RegisterTools(Generic[W]):
    _word_width_: int
    """
    Number of bits in a register word.
    All reads and writes are performed using data vectors of this width.
    All reads and writes are word aligned.
    """

    _addr_unit_width_: int = 8
    """
    `_word_width_` must be an integer multiple of this value.
    """

    _word_stride_: int
    """
    Number of addressable units per word.
    Calculated from `_word_width_` and `_addr_unit_width_`.
    """

    @classmethod
    def _as_word_addr_(cls, unit_addr: Unsigned):
        """
        converts a byte address to a word address

        for a system with 4-Byte words this removes
        the two lsbs.
        """

    @classmethod
    def _as_unit_addr_(cls, word_addr: Unsigned):
        """
        converts a word address to a byte address

        for a system with 4-Byte words this adds
        two zero lsbs.
        """

    class Access(enum.Enum):
        na = enum.auto()
        r = enum.auto()
        w = enum.auto()
        w1 = enum.auto()
        rw = enum.auto()
        rw1 = enum.auto()

    class HwAccess(enum.Enum):
        na = enum.auto()
        r = enum.auto()
        w = enum.auto()
        rw = enum.auto()

    class RegisterObject(std.Template[GenericArg]):
        """

        Customization points:

        _config_
            Used to initialize the RegisterObject.
            Needed because __init__ is reserved for internal use
            by the register implementation.

        _impl_
        _impl_sequential_
        _impl_concurrent_
            Used to implement the behavior of the RegisterObject.

        _basic_read_
        _basic_write_
            These methods connect the register implementation defined
            above to the generic register interface.
        """

        _register_tools_: type[RegisterTools]
        _global_offset_: int
        _word_count_: int

        _readable_: bool = True
        _writable_: bool = True

        def _global_word_offset_(self) -> int: ...
        def _implement_synthesizable_contexts_(self, ctx: SequentialContext):
            """
            This method is used by the address bus implementation to
            instantiate the register object.

            It should never be called from user code.
            """

        def _config_(self, *args, **kwargs):
            """
            Classes derived from `RegisterObject` can override this method to configure
            objects after they have been initialized. If a `RegisterObject` overrides this method
            but does not call it during `AddrMap._config_` it will be called without
            any arguments to provide a default configuration.
            """

        def _impl_(self, ctx: std.SequentialContext) -> None:
            """
            Called by the bus master to implement register object behavior.
            `ctx` is the context of the bus clock.
            """

        def _impl_sequential_(self) -> None:
            """
            Called by the bus master to implement register object behavior.
            """

        def _impl_concurrent_(self) -> None:
            """
            Called by the bus master to implement register object behavior.
            """

        def __init__(self, parent: RegisterTools.RegFile, name: str): ...
        def _flatten_(self) -> list[RegisterTools.RegisterObject]:
            """
            Returns all register objects contained in a register device.
            """

        def _print_(self, *args, **kwargs): ...
        @classmethod
        def _unit_count_(cls) -> int:
            """
            Returns the number of addressable units (bytes) in a register word.
            """

        @classmethod
        def _word_width_(cls) -> int:
            """
            Returns the number of bits in a register word.
            """

        def _contains_addr_(self, addr: Unsigned) -> bool:
            """
            Returns True when the given addr is covered by this register object.
            """

        def _basic_read_(self, addr: Unsigned, meta) -> BitVector:
            """
            This method is called by the bus master for every read access
            to an address in the range covered by this object.

            Can be overwritten with a function or coroutine.

            `addr` is the address used by the hardware read request.

            `meta` exists to pass bus specific meta data.
            Currently always set to None.

            The returned BitVector is used for the read response.
            """

        def _basic_write_(
            self, addr: Unsigned, data: BitVector, mask: std.Mask, meta
        ) -> None:
            """
            This method is called by the bus master for every write access
            to an address in the range covered by this object.

            Can be overwritten with a function or coroutine.

            `addr` is the address used by the hardware write request.

            `data` one word of data received as part of the write request.

            `mask` can be applied to `data` for writes that affect
            only part of a word.

            `meta` exists to pass bus specific meta data.
            Currently always set to None.
            """

        def _info_(self): ...
        def __init_subclass__(cls, readonly=False, writeonly=False): ...

    class RegFile(RegisterObject):
        _member_types_: dict[str, type[RegisterTools[W].RegisterObject]]

        def __init_subclass__(
            cls, *, word_count, readonly=False, writeonly=False
        ) -> None: ...

    class AddrMap(RegFile[0]):
        def __init__(self, *args, **kwargs): ...

    class GenericRegister(RegisterObject):
        """
        Represents a single register word.
        Users can derive from this class an customize the
        read and write behavior by overriding
        the methods `_on_read_` or `_on_write_`.
        """

        _word_count_ = 1

        def _on_read_(self) -> BitVector:
            """
            Customization point for the read behavior of the register.
            By default `cohdl.Null` is returned.

            This method can be overridden with a function or an async-coroutine.
            """

        def _on_write_(self, data: BitVector, mask: std.Mask):
            """
            Customization point for the write behavior of the register.
            By default the received value is discarded.

            This method can be overridden with a function or an async-coroutine.
            """

    class Word(GenericRegister):
        """
        Represents a single register word.
        Similar to register fields, bus write accesses to a Word
        have no effect. Use MemWord to create a writeable version.
        """

        raw: Signal[BitVector]
        def _config_(self, default=Null):
            """
            optional method to configure the default value of the Word
            """

        def __bool__(self) -> bool: ...
        @property
        def next(): ...
        @next.setter
        def next(self, src) -> None: ...
        def __ilshift__(self, src):
            return self

        def val(self):
            return self.raw

    class UWord(Word):
        """
        same as `Word` but the contained raw is unsigned
        """

        raw: Signal[Unsigned]

    class SWord(Word):
        """
        same as `Word` but the contained raw is signed
        """

        raw: Signal[Unsigned]

    class MemWord(Word):
        """
        writeable version of `Word`
        """

    class MemUWord(UWord):
        """
        writeable version of `UWord`
        """

    class MemSWord(SWord):
        """
        writeable version of `SWord`
        """

    #
    #
    #

    class Register(GenericRegister):
        """
        High level abstraction of a single device register word.

        Derive from this class and use the field types described
        below to conveniently access bit fields within the register.

        >>>
        >>> # Defines a register that performs a bitwise or operation
        >>> # of its two low bytes and returns the result via the high byte.
        >>> class OrField(reg32.Register):
        >>>     inp_a: reg32.MemField[7:0]      # MemFields store written values
        >>>     inp_b: reg32.MemField[15:8]     #
        >>>     out:   reg32.Field[31:24]       # Fields to not automatically hold values
        >>>                                     # but can be assigned like signals
        >>>
        >>>     def _config_(self):
        >>>         # optional _config_ method used to
        >>>         # set default value of inp_b to '11111111'
        >>>         self.inp_b._config_(cohdl.Full)
        >>>
        >>>     def _impl_concurrent_(self):
        >>>         self.out <<= self.inp_a.val() | self.inp_b.val()
        >>>
        >>> # Example for FlagField
        >>> # Performs a long running task every time bit 16 is set to '1'.
        >>> # the bit will stay high until the task is done.
        >>> # The api is modeled after std.SyncFlag.
        >>> class FlagExample(reg32.Register):
        >>>     inp:  reg32.MemField[15:0]
        >>>     flag: reg32.FlagField[16]
        >>>
        >>>     async def _impl_sequential_(self):
        >>>         # wait until the flag is set
        >>>         # and clear it once the task is done
        >>>         async with self.flag:
        >>>             await some_long_running_task(self.inp.val())
        >>>
        >>> class CustomReadWrite(reg32.Register):
        >>>     inc_on_read:  reg32.MemUField[7:0]
        >>>     inc_on_write: reg32.MemUField[15:8]
        >>>     normal_mem:   reg32.MemField[31:16]
        >>>
        >>>     def _read_(self):
        >>>         # increments read counter and returns old value
        >>>         self.inc_on_read <<= self.inc_on_read.val() + 1
        >>>         return self
        >>>
        >>>     def _write_(self, value: CustomReadWrite):
        >>>         # only change `inc_on_write`
        >>>         # `inc_on_read` will keep its old value
        >>>         return self(
        >>>             inc_on_write=self.inc_on_write.val() + 1,
        >>>             normal_mem=value.normal_mem
        >>>         )
        >>>
        >>>     # Alternative implementation of _write_ with the same effect
        >>>     # the call operator returns a copy of self with
        >>>     # individual members replaced by objects constructed from the
        >>>     # provided values.
        >>>     def _write_(self, value: CustomReadWrite):
        >>>         # only change `inc_on_write`
        >>>         # `inc_on_read` will keep its old value
        >>>         return value(
        >>>             inc_on_read=self.inc_on_read,
        >>>             inc_on_write=self.inc_on_write.val() + 1
        >>>         )
        >>>
        """

        def __ilshift__(self: Self, source: Self):
            """
            `RegisterObjects` support direct signal assignments.
            """
            return self

        @property
        def next(self) -> NoReturn:
            """
            `next` is a write only property.
            """

        @next.setter
        def next(self: Self, value: Self):
            """
            alternative to the `<<=` operator.
            """

        def _config_(self, default=Null):
            """
            Used provide default values for `MemFields`.
            """

        def __init_subclass__(cls, readonly=None, writeonly=None): ...
        def _on_read_(self: Self) -> Self:
            """
            The return value is serialized into a BitVector and
            sent as part of the read response.
            """

        def _on_write_(self: Self, data: Self) -> Self | None:
            """
            The return value is used to update all `MemFields` and `FlagFields`
            in the register. If no such field exists the return value
            is ignored and may be set to None.
            """

        def __call__(self: Self, **kwargs: Any) -> Self:
            """
            Returns a copy of `self`, the optional keyword arguments
            are used to replace members of the result with `FieldObjects`
            constructed from the provided value.
            """

    class Field:
        """
        Represents a range of one or more bits in a `Register`.

        Declaring a `Field` has no side effect. It only serves as
        a convenient reference to register bits.
        """

        @classmethod
        def _count_bits_(cls) -> int:
            """
            Part of the standard serialization api.
            See `std.count_bits`.
            """

        @classmethod
        def _from_bits_(cls: type[Self], val: BitVector, qualifier) -> Self:
            """
            Part of the standard serialization api.
            See `std.from_bits`.
            """

        def _to_bits_(self) -> BitVector:
            """
            Part of the standard serialization api.
            See `std.to_bits`.
            """

        def __ilshift__(self, other):
            return self

        def __ixor__(self, other):
            return self

        def __init__(self, value, _qualifier_=Signal): ...
        def val(self):
            """
            returns the current value of the `Field`
            """

    class UField(Field):
        """
        A specialization of `Field` that represents an unsigned
        value instead of a BitVector.
        """

        def __init__(self, value: Unsigned, _qualifier_=Signal): ...
        def val(self) -> Unsigned: ...

    class SField(Field):
        """
        A specialization of `Field` that represents a signed
        value instead of a BitVector.
        """

        def __init__(self, value: Signed, _qualifier_=Signal): ...
        def val(self) -> Signed: ...

    class MemField(Field):
        """
        `MemField` is functionally equivalent to the normal `Field` type.
        The only difference is, that values written to this field are stored.
        """

    class MemUField(UField):
        """
        A specialization of `MemField` that represents an unsigned
        value instead of a BitVector.
        """

    class MemSField(SField):
        """
        A specialization of `MemField` that represents a signed
        value instead of a BitVector.
        """

    class FlagField:
        """
        A single bit field that can be set by the bus master
        and cleared by the cohdl implementation.
        """

        @classmethod
        def _count_bits_(cls) -> int:
            """
            Part of the standard serialization api.
            See `std.count_bits`.
            """

        @classmethod
        def _from_bits_(
            cls, val: BitVector[1], qualifier=std.Value
        ) -> RegisterTools.FlagField:
            """
            Part of the standard serialization api.
            See `std.from_bits`.
            """

        def _to_bits_(self) -> BitVector[1]:
            """
            Part of the standard serialization api.
            See `std.to_bits`.
            """

        def flag(self) -> SyncFlag:
            """
            returns the interal `std.SyncFlag` object
            """

        def set(self) -> None:
            """
            Set the flag to True, usually not called from user
            code. Instead this method is called then a '1' bit is written
            to the corresponding register bit.
            """

        def clear(self) -> None:
            """
            Set the flag to False.
            """

        def is_set(self) -> bool:
            """
            Returns True when the flag is set.
            """

        def is_clear(self) -> bool:
            """
            Returns False when the flag is set.
            """

        async def __aenter__(self): ...
        async def __aexit__(self, val, type, traceback): ...

    class MetaNotifyBase(type):
        @property
        def Read(cls):
            return cls

        @property
        def Write(cls):
            return cls

    class NotifyBase(metaclass=MetaNotifyBase):
        """
        Register notifications are used to inform HDL code, that a
        read or write operations was performed on a register.

        Like fields, notifications are defined using annotations on a register type.
        Unlike fields, notifications are not associated with register bits.
        The class properties `Read` and `Write` are used to select, when
        the notification should be triggered.

        Checkout the examples of `PushOnNotify` and `FlagOnNotify`.
        """

        def notify(self):
            """
            called when the selected event (read or write) occurs on the register
            """

        def __bool__(self) -> bool:
            """
            returns true when a notification is pending
            """

    class PushOnNotify(NotifyBase):
        """
        Defines a single bit signal, that is set to true for one clock cycle
        when the corresponding event (read or write) occurs on the register.

        >>> class ExamplePush(reg32.Register):
        >>>     data: reg32.MemField[15:0, Null]
        >>>     rd_cnt: reg32.UField[23:16, Null]
        >>>     wr_cnt: reg32.UField[31:14, Null]
        >>>
        >>>     rd_notification: reg32.PushOnNotify.Read   # use .Read/.Write to select the monitored
        >>>     wr_notification: reg32.PushOnNotify.Write  # type of event
        >>>
        >>>     def _impl_sequential_(self):
        >>>         if self.rd_notification:
        >>>             # self.rd_notification is true for one clock cycle after each
        >>>             # read from the current register
        >>>             self.rd_cnt <<= self.rd_cnt.val() + 1
        >>>         if self.wr_notification:
        >>>             # self.wr_notification is true for one clock cycle after each
        >>>             # write to the current register
        >>>             self.wr_cnt <<= self.wr_cnt.val() + 1
        """

    class FlagOnNotify(NotifyBase):
        """
        Defines a std.SyncFlag that is set when the corresponding event (read or write)
        occurs on the register.

        >>> class ExampleFlag(reg32.Register):
        >>>     data: reg32.MemField[15:0, Null]
        >>>     rd_cnt: reg32.UField[23:16, Null]
        >>>
        >>>     rd_notification: reg32.FlagOnNotify.Read
        >>>
        >>>     async def _impl_sequential_(self):
        >>>         # wait for notification, then increment
        >>>         # the read counter register
        >>>         async with self.rd_notification:
        >>>             self.rd_cnt <<= self.rd_cnt.val() + 1
        """

        def clear(self):
            """
            reset the flag
            """

        async def __aenter__(self):
            """
            wait until the notification is received
            """

        async def __aexit__(self, val, type, traceback):
            """
            clear the notification
            """

    class Input(GenericRegister, readonly=True):
        """
        A single read-only register that connects a Signal to the address bus.

        >>>
        >>> # A register device, that provides access to hardware buttons
        >>> # and LEDs.
        >>> class ExampleInOut(reg32.RegFile, word_count=2):
        >>>     reg_btn: reg32.Input[0x0]   # input register at offset 0x00
        >>>     reg_led: reg32.Output[0x4]  # output register at offset 0x04
        >>>
        >>>     def _config_(self, signal_button, signal_led):
        >>>         # connect signal_button to the register
        >>>         self.reg_btn._config_(signal_button)
        >>>         # connect signal_led to the register
        >>>         self.reg_led._config_(signal_led)
        """

        def _config_(
            self,
            signal: Signal[BitVector],
            /,
            offset: int | None = None,
            padding: int | None = None,
            lsbs: bool | None = False,
            msbs: bool | None = False,
        ):
            """
            When the width of the
            """

    class Output(GenericRegister, writeonly=True):
        """
        A single write-only register, that connects a Signal to the address bus.

        See example of `Input`.
        """

        def _config_(self, signal: Signal[BitVector]) -> None: ...

    class Array(RegisterObject):
        """
        A range of register objects of the same type located at equidistant address locations.
        """

        def __len__(self) -> int: ...
        def __getitem__(self, index: int) -> RegisterTools.RegisterObject: ...

    class AddrRange(RegisterObject):
        def _on_read_(self, addr: Unsigned) -> BitVector:
            """
            called on read accesses to the address range

            addr is the global address within the bus address space
            """

        def _on_write_(self, addr: Unsigned, data: BitVector, mask: std.Mask):
            """
            called on write accesses to the address range

            addr is the global address within the bus address space
            """

        def _on_read_relative_(self, addr: Unsigned):
            """
            Equivalent to _on_read_ but the given address is
            relative to the base address of this object.

            Only one of `_on_read_` and `_on_read_relative_` can be specified.
            """

        def _on_write_relative_(self, addr: Unsigned, data: BitVector, mask: std.Mask):
            """
            Equivalent to _on_write_relative_ but the given address is
            relative to the base address of this object.

            Only one of `_on_write_`and `_on_write_relative_` can be specified.
            """

    class Memory(AddrRange):
        """

        >>> # Example
        >>>
        >>> class MyAddrMap(reg32.AddrMap, word_count=256):
        >>>     prog_mem: RoMemory[0x000:0x100]
        >>>     ram:      Memory[0x200:400]
        >>>
        >>>     # optional configuration for the memory objects
        >>>     def _config_(self, prog_mem: list[BitVector[32]]):
        >>>         # define the initial content of the program memory
        >>>         prog_mem._config_(initial=prog_mem)
        >>>
        >>>         # For memory regions, that are allowed to keep their state
        >>>         # after a reset the `noreset` argument can be set to True.
        >>>         #
        >>>         # (in this example it does not matter because no initial value is provided)
        >>>         #
        >>>         # For memory regions, that do not require sub-word write access
        >>>         # `mask_mode` can be set to `IGNORE` to remove the otherwise
        >>>         # generated read-modify-write logic.
        >>>         ram._config_(
        >>>             noreset=True,
        >>>             mask_mode=reg32.Memory.MaskMode.IGNORE
        >>>         )
        """

        class MaskMode(enum.Enum):
            """
            defines, how memory write operations are implemented
            """

            IMMEDIATE = enum.auto()
            """
            Default mask Mode, memory writes are implemented as read-modify-write
            operations in a single clock cycle.

            >>> # pseudo code for write operation
            >>> # write is implemented as a read-modify-write operation
            >>> # in a single clock cycle
            >>> mem_array[index] <<= (mem_array[index] & ~mask) | (data & mask)
            """
            IGNORE = enum.auto()
            """
            In this mode, the mask parameter is ignored. This mode requires the least FPGA
            resources and should be used whenever sub-word write access is not needed.
            
            >>> # pseudo code for write operation
            >>> # data is directly assigned to memory word
            >>> mem_array[index] <<= data
            """
            READBACK = enum.auto()
            """
            This mode splits the read-modify-write operation of each memory
            write into two clock cycles.

            >>> # pseudo code for write operation
            >>> old_data = mem_array[index]
            >>> await std.wait_for(1)
            >>> mem_array[index] <<= (old_data & ~mask) | (data & mask)
            """
            SPLIT_WORDS = enum.auto()
            """
            Mode SPLIT_WORDS changes the generated memory layout.
            Instead of a single memory array of word width, each addressable unit (byte)
            within a word is stored in a separate array.
            The advantage of this mode is, that the memory write process
            can modify individual bytes without a separate read logic.

            >>> # pseudo code for write operation
            >>> # each byte is assigned individually
            >>>
            >>> if mask[0]:
            >>>     mem_array_0[index] <<= data[7:0]
            >>> if mask[1]:
            >>>     mem_array_1[index] <<= data[15:8]
            >>> if mask[2]:
            >>>     mem_array_2[index] <<= data[23:16]
            >>> if mask[3]:
            >>>     mem_array_3[index] <<= data[31:16]
            """

        def _config_(
            self,
            initial: Null | Full | list[BitVector] | None = None,
            noreset: bool = False,
            mask_mode: MaskMode = MaskMode.IMMEDIATE,
            allow_unaligned: bool = False,
            inline: bool = False,
        ):
            """
            Configure how the memory object is initialized and implemented.

            `initial` defines the initial memory content.

            When `noreset` is set to True the memory content is not restored
            when the register system is reset. This should be set whenever
            possible, since the reset logic prevents block ram inference.

            `mask_mode` defines how write operations are implemented.

            Only aligned reads/writes are allowed unless `allow_unaligned` is set to True.
            This is only possible with `MaskMode.SPLIT_WORDS`.

            When `inline` is set to True the read/write memory accesses are performed
            in the context of the bus transaction.
            This reduces the read latency by one clock cycle but prevents
            block RAM inference.
            """

    class RoMemory(Memory):
        """
        Specialization of `Memory` that does not implement write operations.
        """

        def _config_(
            self,
            initial,
            mask_mode: RegisterTools.Memory.MaskMode = RegisterTools.Memory.MaskMode.IMMEDIATE,
            allow_unaligned: bool = False,
            inline=False,
        ):
            """
            `initial` defines the initial memory content. This is a mandatory argument
            because it is the only way to load data into the read-only memory.

            `mask_mode` mainly affects the masking behavior during writes.
            It can be usually be ignored for read-only memory. However when `allow_unaligned`
            is set to True `MaskMode.SPLIT_WORDS` is required.
            """

reg32 = RegisterTools[32, 8]
"""
Provides register tools pre-configured with a
word size of 32 bits and 8 bits per byte.
"""

Access = RegisterTools.Access
HwAccess = RegisterTools.HwAccess



from __future__ import annotations

from typing import Callable, TypeVar, Generic

S = TypeVar("S")
T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")


class IdMap(Generic[S, T], dict[int, T]):
    @staticmethod
    def merge(
        *sources: IdMap[U, V], on_conflict: Callable | None = None
    ) -> IdMap[U, V]:
        result: dict[int, U] = {}

        if on_conflict is None:
            on_conflict = lambda a, b: a

        for source in sources:
            conflict_keys = result.keys() & source.keys()

            result.update(
                {
                    key: dict.__getitem__(source, key)
                    for key in source.keys() - conflict_keys
                }
            )

            result.update(
                {
                    key: (
                        dict.__getitem__(source, key)
                        if result[key] is dict.__getitem__(source, key)
                        else on_conflict(result[key], dict.__getitem__(source, key))
                    )
                    for key in conflict_keys
                }
            )

        return IdMap(result)

    def map_self(self, obj):
        self[obj] = obj

    def __getitem__(self, key_obj: S | int) -> T:
        if type(key_obj) is int:
            return super().__getitem__(key_obj)
        return super().__getitem__(id(key_obj))

    def __setitem__(self, key_obj: S | int, value: T) -> None:
        assert type(key_obj) is not str

        if type(key_obj) is int:
            return super().__setitem__(key_obj, value)
        return super().__setitem__(id(key_obj), value)

    def __delitem__(self, key_obj: S | int) -> None:
        if type(key_obj) is int:
            super().__delitem__(key_obj)
        else:
            super().__delitem__(id(key_obj))

    def __contains__(self, key_obj: S | int) -> bool:
        if type(key_obj) is int:
            return key_obj in dict.keys(self)
        return super().__contains__(id(key_obj))


class IdSet(Generic[T]):
    def __init__(self, *, _content: IdMap | None = None) -> None:
        self._content: IdMap[T, T] = IdMap() if _content is None else _content

    def __and__(self, other: IdSet[T]) -> IdSet[T]:
        keys = self._content.keys() & other._content.keys()
        return IdSet(_content=IdMap({key: self._content[key] for key in keys}))

    def __iter__(self):
        return iter(self._content.values())

    def add(self, element: T):
        self._content[element] = element

    def __contains__(self, element: T):
        return element in self._content

    def __len__(self) -> int:
        return len(self._content)

    def union(*others: IdSet[S]) -> IdSet[T | S]:
        return IdSet(_content=IdMap.merge(*[o._content for o in others]))

    def clear(self):
        self._content.clear()

    def update(self, *others):
        for o in others:
            for elt in o:
                self.add(elt)



from __future__ import annotations

from cohdl import std
from cohdl import Unsigned, BitVector, Bit, Null, Full, Signal, Entity, Port

from cohdl.std import reg
from cohdl.std.axi.axi4_light import Axi4Light

from cohdl.std.reg import reg32


class OrExample(reg32.Register):
    a: reg32.MemField[7:0]
    b: reg32.MemField[15:8]
    c: reg32.MemField[23:16]

    result: reg32.Field[31:24]

    irq_rd: reg32.ReadTick
    irq_wr: reg32.WriteTick

    def _on_read_(self) -> BitVector:
        std.comment("THIS IS AN OrExample")
        self.a = 3
        return self(a=Null, b=Full)

    def _impl_concurrent_(self):
        self.result <<= self.a.value() | self.b.value() & self.c.value()


class FlagExample(reg32.Register):
    """
    a flag example register
    """

    a: reg32.MemField[7:0]
    b: reg32.MemField[15:8]

    flag: reg32.FlagField[20]

    result: reg32.Field[31:24]

    async def _impl_sequential_(self):
        await std.wait_for(100)
        self.flag.clear()

    def _impl_concurrent_(self):
        self.result <<= self.a.value() | self.b.value()


class SubDevice(reg32.RegisterDevice, word_count=8):
    reg_1: OrExample[0]
    reg_2: OrExample[4]
    reg_3: FlagExample[8]
    reg_4: FlagExample[12]


class RgbReg(reg32.Register):
    """
    rgb fhgjfhg
    """

    r: reg32.MemUField[7:0]
    g: reg32.MemUField[15:8]
    b: reg32.MemUField[23:16]
    cnt: reg32.UField[31:24]

    def _config_(self, r, g, b):
        self.sig_r = r
        self.sig_g = g
        self.sig_b = b
        self.counter = Signal[Unsigned[8]](0)

    async def _impl_sequential_(self):
        await std.wait_for(1000)
        self.counter <<= self.counter + 1
        self.sig_r <<= self.counter < self.r.value()
        self.sig_g <<= self.counter < self.g.value()
        self.sig_b <<= self.counter < self.b.value()
        self.cnt <<= self.counter


class MyAddrRange(reg32.AddrRange, word_count=64):
    def _config_(self):
        self._mem = Signal[BitVector[32]](name="ASDFASDF_RANGE")

    def _on_read_relative_(self, addr: Unsigned):
        return self._mem

    def _on_write_relative_(self, addr: Unsigned, data: BitVector, mask: BitVector):
        self._mem <<= data


class RootDevice(reg32.RootDevice, word_count=1024):
    sub_a: SubDevice[0]
    sub_b: SubDevice[32]

    rgb: RgbReg[64]

    my_range: MyAddrRange[256]

    my_array: reg32.Array[OrExample, 512:576:8]

    def _config_(self, r, g, b):
        self.rgb._config_(r, g, b)


class MyEntity(Entity):
    clk = Port.input(Bit)
    reset = Port.input(Bit)

    r = Port.output(Bit)
    g = Port.output(Bit)
    b = Port.output(Bit)

    def architecture(self):
        clk = std.Clock(self.clk)
        reset = std.Reset(self.reset)

        axi = Axi4Light.signal(clk, reset, 32, 32)

        axi.connect_addr_map(RootDevice(self.r, self.g, self.b))


with open("example.vhdl", "w") as file:
    print(std.VhdlCompiler.to_string(MyEntity), file=file)



from cohdl import pyeval

from ._builtins import vhdl


class _SeqNode:
    @staticmethod
    def write_node(node):
        if isinstance(node, _SeqNode):
            return node.write()
        else:
            return f"{vhdl:{node!r}}"

    def write(self): ...


_write_node = _SeqNode.write_node


class When(_SeqNode):
    @pyeval
    def __init__(self, precond, postcond=None, immediate=True):
        self.precond = precond
        self.postcond = postcond
        self.immediate = immediate

    @pyeval
    def then(self, state, times=None):
        if self.postcond is None:
            self.immediate = True
            self.postcond = State(state, times)
        elif isinstance(self.postcond, Sequence):
            self.postcond.then(state, times)
        else:
            self.postcond = Sequence(self.postcond).then(state, times)
        return self

    @pyeval
    def next(self, state, times=None):
        if self.postcond is None:
            self.immediate = False
            self.postcond = State(state, times)
        elif isinstance(self.postcond, Sequence):
            self.postcond.next(state, times)
        else:
            self.postcond = Sequence(self.postcond).next(state, times)
        return self

    def write(self):
        if self.immediate:
            return (
                f"{vhdl:({_write_node(self.precond)} |-> {_write_node(self.postcond)})}"
            )
        else:
            return (
                f"{vhdl:({_write_node(self.precond)} |=> {_write_node(self.postcond)})}"
            )


class State(_SeqNode):
    def __init__(self, cond, times=None, consecutive=True):
        self.cond = cond
        self.times = times
        self.consecutive = consecutive

    def write_times(self):
        start_char = "*" if self.consecutive else "="

        if self.times is None:
            return ""
        elif isinstance(self.times, tuple):
            if len(self.times) == 0:
                return f"{vhdl:[*]}"
            elif len(self.times) == 1:
                raise AssertionError("not implemented")
            elif len(self.times) == 2:
                return f"{vhdl:[{start_char}{self.times[0]} to {self.times[1]}]}"
            else:
                raise AssertionError("")
        else:
            return f"{vhdl:[->{self.times}]}"

    def write(self):
        return f"{vhdl:{_write_node(self.cond)}{self.write_times()}}"


class Next(State):
    def __init__(self, cond, times=None, consecutive=True):
        super().__init__(cond, times, consecutive)

    def write(self):
        return f"{vhdl: ; {_write_node(self.cond)}{self.write_times()}}"


class Then(State):
    def __init__(self, cond, times=None, consecutive=True):
        super().__init__(cond, times, consecutive)

    def write(self):
        return f"{vhdl: : {_write_node(self.cond)}{self.write_times()}}"


class Wait(State):
    def __init__(self, times=None):
        super().__init__(None, times)

    def write(self):
        return f"{vhdl: ; true {self.write_times()}}"


class Sequence(_SeqNode):
    @staticmethod
    def _write_seq(nodes: list):
        if len(nodes) == 0:
            return ""
        else:
            first, *rest = nodes
            return f"{vhdl:{_write_node(first)}{Sequence._write_seq(rest)}}"

    def __init__(self, start, *seq):
        if isinstance(start, _SeqNode):
            self.start = start
        else:
            self.start = State(start)

        self.seq = [*seq]

    @pyeval
    def then(self, state, times=None, consecutive=True):
        self.seq.append(Then(state, times=times, consecutive=consecutive))
        return self

    @pyeval
    def next(self, state, times=None, consecutive=True):
        self.seq.append(Next(state, times=times, consecutive=consecutive))
        return self

    @pyeval
    def wait(self, times):
        self.seq.append(Wait(times))
        return self

    def write(self):
        return f"{vhdl:<%{_write_node(self.start)}{self._write_seq(self.seq)}%>}"



from __future__ import annotations

from abc import abstractmethod
import enum
import typing

from typing import Callable, Tuple, cast

from cohdl._core._type_qualifier import (
    Signal,
    Variable,
    Temporary,
    Port,
    Generic,
    TypeQualifier,
    Offset,
    Slice,
)

from cohdl._core._bit import Bit
from cohdl._core._intrinsic import (
    _BitSignalEvent,
    _BitSignalEventGroup,
)

from cohdl.utility import IndentBlock, IdMap, IdSet
from cohdl.utility.source_location import SourceLocation
from cohdl.utility.virtual_traceback import VirtualFrame

from cohdl._core._intrinsic import _SensitivitySpec, _SensitivityAll, _SensitivityList
from cohdl._core._primitive_type import is_primitive
from cohdl._core import enum as cohdl_enum
from cohdl._core._context import EntityInfo

from cohdl._core import _InlineCode
from cohdl._core import _intrinsic_operations as intr_op


class VisitException(Exception):
    def __init__(self, src_statement: Statement, original: Exception):

        if isinstance(original, VisitException):
            self.with_traceback(original.original.__traceback__)
            self.src_statement = original.src_statement
            self.original = original.original
        else:
            self.with_traceback(original.__traceback__)
            self.src_statement = src_statement
            self.original = original


class AccessFlags(enum.Flag):
    _ZERO = 0
    READ = enum.auto()
    WRITE = enum.auto()
    PUSH = enum.auto()

    def short_repr(self) -> str:
        ret = ""

        def field(flag: AccessFlags, letter: str):
            nonlocal ret

            ret += letter if flag & self else "-"

        field(AccessFlags.READ, "r")
        field(AccessFlags.WRITE, "w")
        field(AccessFlags.PUSH, "p")

        return ret

    def is_read(self):
        return bool(self & AccessFlags.READ)

    def is_pushed(self):
        return bool(self & AccessFlags.PUSH)

    def is_written(self):
        return bool(self & AccessFlags.WRITE)


def _visit_referenced_objects(self, operation):
    # used to include objects in refspec in visited_objects

    def visit_single_object(obj, access):
        if isinstance(obj, TypeQualifier):
            for ref in obj._ref_spec:
                if isinstance(ref, Offset):
                    if isinstance(ref.offset, TypeQualifier):
                        ref.offset = operation(ref.offset, AccessFlags.READ)

                elif isinstance(ref, Slice):
                    if isinstance(ref.start, TypeQualifier):
                        ref.start = operation(ref.start, AccessFlags.READ)
                    if isinstance(ref.stop, TypeQualifier):
                        ref.stop = operation(ref.stop, AccessFlags.READ)
                else:
                    raise AssertionError("invalid ref_spec")

        return operation(obj, access)

    return self.visit_objects(visit_single_object)


class Statement:
    _current_frame = None

    @staticmethod
    def _obj_str(obj) -> str:
        return f"id: {id(obj):#x}, {obj}"

    def __init__(self):
        self._parent_state: _State | None = None
        self._frame: VirtualFrame = Statement._current_frame

    def visit(self, operation):
        try:
            return operation(self)
        except Exception as err:
            raise VisitException(self, err)

    @abstractmethod
    def visit_objects(
        self, operation: Callable[[typing.Any, AccessFlags], typing.Any]
    ): ...

    visit_referenced_objects = _visit_referenced_objects

    def set_parent_state(self, state: _State):
        self._state = state

    def get_parent_state(self) -> _State:
        return self._state

    @abstractmethod
    def update_transitions(self, state_map: IdMap):
        """
        point target of all transitions in the statement to new state defined in state_map
        """
        raise AssertionError("abstract method called")

    @abstractmethod
    def copy(self) -> Statement:
        raise AssertionError("abstract method called")

    @abstractmethod
    def dump(self) -> IndentBlock: ...


class Expression(Statement):
    def __init__(
        self,
        result,
    ):
        self._result = result
        super().__init__()

    def result(self):
        return self._result

    @abstractmethod
    def copy(self) -> Expression:
        raise AssertionError("abstract method called")

    @abstractmethod
    def dump(self) -> IndentBlock: ...

    @abstractmethod
    def visit_objects(self, operation):
        try:
            self._result = operation(self._result, AccessFlags.WRITE)
        except Exception as err:
            raise VisitException(self, err)

    def update_transitions(self, state_map: IdMap):
        pass


class CodeBlock(Statement):
    def __init__(self, content: list[Statement], parent: CodeBlock | None):
        if None in content:
            raise AssertionError("ERROR")

        self._content: list[Statement] = []
        self._parent = parent

        self._level: int = 0 if parent is None else parent._level + 1
        self._root: CodeBlock = self if parent is None else parent._root

        for stmt in content:
            if isinstance(stmt, CodeBlock):
                self._content.extend(stmt._content)
            else:
                self._content.append(stmt)

        super().__init__()

    def _fix_alias(self):
        # replace all reads from locally constructed signals
        # with reads from a temporary

        alias_map = IdMap()

        def collect_alias(node):
            if isinstance(node, _SignalAlias):
                alias_map[node.signal] = node.replacement
                return Nop()
            return node

        self.visit(collect_alias)

        def apply_alias(obj, access: AccessFlags):
            # replace all reads from aliased signals with replacement
            if access is AccessFlags.READ and hasattr(obj, "_root"):
                if obj._root in alias_map:
                    return Temporary[obj.type](
                        obj._value, _root=alias_map[obj._root], _ref_spec=obj._ref_spec
                    )
            return obj

        self.visit_objects(apply_alias)

    def get_parent_block(self, level):
        assert (
            self._level >= level
        ), "internal error: wrong nesting order of code blocks"

        if self._level == level:
            return self

        assert self._parent is not None, "internal error: code block has no parent"
        return self._parent.get_parent_block(level)

    @classmethod
    def common_block(cls, blocks: list[CodeBlock]):
        """
        check, if a and b are sub-blocks of a common parent
        return common parent, if it exists
        otherwise return None
        """

        if len(blocks) == 0:
            return None

        min_level = min(block._level for block in blocks)

        blocks = [block.get_parent_block(min_level) for block in blocks]

        level = min_level

        while True:
            first, *rest = blocks
            if all(block is first for block in rest):
                return first

            if level == 0:
                return None
            level -= 1

            blocks = [block._parent for block in blocks]

    def visit(self, operation):
        try:
            self._content = [stmt.visit(operation) for stmt in self._content]
            return operation(self)
        except Exception as err:
            raise VisitException(self, err)

    def visit_objects(self, operation: Callable):
        try:
            for stmt in self._content:
                stmt.visit_objects(operation)
        except Exception as err:
            raise VisitException(self, err)

    def update_transitions(self, state_map: IdMap):
        for stmt in self._content:
            stmt.update_transitions(state_map)

    # TODO: decide, if parent argument required,
    # breaks symmetry with other statements
    # maybe set parent of copy to None and set parent when
    # when adding the copy to the parent codeblock
    def copy(self, parent: CodeBlock | None) -> CodeBlock:
        return CodeBlock(
            [
                stmt.copy(self) if isinstance(stmt, CodeBlock) else stmt.copy()
                for stmt in self._content
            ],
            parent,
        )

    def empty(self) -> bool:
        # comments are considered empty so they don't affect
        # the special case where an await expression is the first
        # statement in a sequential context
        return len(self._content) == 0 or all(
            isinstance(elem, Comment) for elem in self._content
        )

    def addfront(self, stmt: Statement):
        if stmt is None:
            raise AssertionError("ERROR")
        self._content.insert(0, stmt)

    def append(self, stmt: Statement):
        if stmt is None:
            raise AssertionError("ERROR")
        self._content.append(stmt)

    def content(self) -> list[Statement]:
        return self._content

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="Codeblock", content=[stmt.dump() for stmt in self._content]
        )


class Nop(Statement):
    def __init__(self):
        super().__init__()

    def visit_objects(self, operation: Callable):
        pass

    def copy(self):
        return Nop()

    def dump(self):
        return IndentBlock(title="Nop", content=[])


class Comment(Statement):
    def __init__(self, lines: list[str] = None):
        super().__init__()
        self.lines = lines if lines is not None else []

    def visit_objects(self, operation: Callable):
        pass

    def copy(self):
        return Comment([*self.lines])

    def dump(self) -> IndentBlock:
        return IndentBlock(title="Comment", content=[*self.lines])


class Event:
    Type = _BitSignalEvent.Type

    def __init__(self, event: _BitSignalEvent):
        self.sig = event.sig
        self.event_type = event.event_type

    def visit(self, operation):
        try:
            return operation(self)
        except Exception as err:
            raise VisitException(self, err)

    def visit_objects(self, operation: Callable):
        try:
            self.sig = operation(self.sig, AccessFlags.READ)
        except Exception as err:
            raise VisitException(self, err)

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="_Event",
            content=[
                IndentBlock(title="type", content=str(self.event_type)),
            ],
        )


class EventGroup:
    Operation = _BitSignalEventGroup.Operation

    def __init__(self, event_grp: _BitSignalEventGroup):
        self.operation = event_grp.operation
        self.events = []

        for event in event_grp.events:
            if isinstance(event, _BitSignalEvent):
                self.events.append(Event(event))
            elif isinstance(event, _BitSignalEventGroup):
                self.events.append(EventGroup(event.operation, event.events))

    def visit(self, operation):
        try:
            self.events = [operation(event) for event in self.events]
            return operation(self)
        except Exception as err:
            raise VisitException(self, err)

    def visit_objects(self, operation: Callable):
        try:
            for event in self.events:
                event.visit_objects(operation)
        except Exception as err:
            raise VisitException(self, err)

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="_EventGroup",
            content=[
                IndentBlock(title="operation", content=str(self.operation)),
                *[event.dump() for event in self.events],
            ],
        )


class If(Statement):
    def __init__(
        self,
        test,
        body: CodeBlock,
        orelse: CodeBlock,
    ):
        if isinstance(test, _BitSignalEvent):
            self._test = Event(test)
        elif isinstance(test, _BitSignalEventGroup):
            self._test = EventGroup(test)
        else:
            self._test = test

        self._body = body
        self._orelse = orelse

        super().__init__()

    def visit(self, operation):
        try:
            if isinstance(self._test, (Event, EventGroup)):
                self._test = self._test.visit(operation)

            self._body = self._body.visit(operation)
            self._orelse = self._orelse.visit(operation)
            return operation(self)
        except Exception as err:
            raise VisitException(self, err)

    def visit_objects(self, operation: Callable):
        try:
            if isinstance(self._test, (Event, EventGroup)):
                self._test.visit_objects(operation)
            else:
                self._test = operation(self._test, AccessFlags.READ)

            self._body.visit_objects(operation)
            self._orelse.visit_objects(operation)
        except Exception as err:
            raise VisitException(self, err)

    def copy(self) -> If:
        return If(self._test, self._body.copy(None), self._orelse.copy(None))

    def update_transitions(self, state_map: IdMap):
        self._body.update_transitions(state_map)
        self._orelse.update_transitions(state_map)

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="If",
            content=[
                IndentBlock(title="test", content=Statement._obj_str(self._test)),
                IndentBlock(title="body", content=self._body.dump()),
                IndentBlock(title="orelse", content=self._orelse.dump()),
            ],
        )


class Compare(Expression):
    Operator = intr_op.ComparisonOperator

    def __init__(
        self,
        op: Compare.Operator,
        lhs,
        rhs,
        result: Temporary[Bit],
    ):
        self._op = op
        self._lhs = lhs
        self._rhs = rhs

        super().__init__(result)

    def visit_objects(self, operation: Callable):
        try:
            self._lhs = operation(self._lhs, AccessFlags.READ)
            self._rhs = operation(self._rhs, AccessFlags.READ)
            self._result = operation(self._result, AccessFlags.WRITE)
        except Exception as err:
            raise VisitException(self, err)

    def copy(self) -> Compare:
        return Compare(self._op, self._lhs, self._rhs, self.result())

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"Compare, op = {self._op}",
            content=[
                f"lhs    = {Statement._obj_str(self._lhs)}",
                f"rhs    = {Statement._obj_str(self._rhs)}",
                f"result = {Statement._obj_str(self.result())}",
            ],
        )


class UnaryOp(Expression):
    Operator = intr_op.UnaryOperator

    def __init__(
        self,
        op: UnaryOp.Operator,
        arg,
        result,
    ):
        assert is_primitive(TypeQualifier.decay(arg))
        assert is_primitive(TypeQualifier.decay(result))
        assert isinstance(op, UnaryOp.Operator)

        self._op = op
        self._arg = arg

        super().__init__(result)

    def visit_objects(self, operation: Callable):
        try:
            self._arg = operation(self._arg, AccessFlags.READ)
            self._result = operation(self._result, AccessFlags.WRITE)
        except Exception as err:
            raise VisitException(self, err)

    def copy(self) -> UnaryOp:
        return UnaryOp(self._op, self._arg, self.result())

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"UnaryOp, op = {self._op}",
            content=[
                f"arg    = {Statement._obj_str(self._arg)}",
                f"result = {Statement._obj_str(self.result())}",
            ],
        )


class BinOp(Expression):
    Operator = intr_op.BinaryOperator

    def __init__(
        self,
        op: BinOp.Operator,
        lhs,
        rhs,
        result,
    ):
        self._op = op
        self._lhs = lhs
        self._rhs = rhs

        super().__init__(
            result,
        )

    def visit_objects(self, operation: Callable):
        try:
            self._lhs = operation(self._lhs, AccessFlags.READ)
            self._rhs = operation(self._rhs, AccessFlags.READ)
            self._result = operation(self._result, AccessFlags.WRITE)
        except Exception as err:
            raise VisitException(self, err)

    def copy(self) -> BinOp:
        return BinOp(self._op, self._lhs, self._rhs, self.result())

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"BinOp, op = {self._op}",
            content=[
                f"lhs    = {Statement._obj_str(self._lhs)}",
                f"rhs    = {Statement._obj_str(self._rhs)}",
                f"result = {Statement._obj_str(self.result())}",
            ],
        )


class All(Expression):
    """
    and operation over arbitrary number of arguments
    """

    def __init__(self, args: list, result):
        self._args = args

        super().__init__(result)

    def visit_objects(self, operation: Callable):
        try:
            self._args = [operation(arg, AccessFlags.READ) for arg in self._args]
            self._result = operation(self._result, AccessFlags.WRITE)
        except Exception as err:
            raise VisitException(self, err)

    def copy(self) -> All:
        return All([*self._args], self.result())

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"All, len = {len(self._args)}",
            content=[
                f"{n}: {Statement._obj_str(arg)}" for n, arg in enumerate(self._args)
            ],
        )


class Any(Expression):
    """
    or operation over arbitrary number of arguments
    """

    def __init__(self, args: list, result):
        self._args = args

        super().__init__(result)

    def visit_objects(self, operation: Callable):
        try:
            self._args = [operation(arg, AccessFlags.READ) for arg in self._args]
            self._result = operation(self._result, AccessFlags.WRITE)
        except Exception as err:
            raise VisitException(self, err)

    def copy(self) -> Any:
        return Any([*self._args], self.result())

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"All, len = {len(self._args)}",
            content=[
                f"{n}: {Statement._obj_str(arg)}" for n, arg in enumerate(self._args)
            ],
        )


class SignalAssignment(Statement):
    def __init__(self, target_signal: Signal, source, frame=None):
        self._target = target_signal
        self._source = source

        super().__init__()

        if frame is not None:
            self._frame = frame

    def visit_objects(self, operation: Callable):
        try:
            self._target = operation(self._target, AccessFlags.WRITE)
            self._source = operation(self._source, AccessFlags.READ)
        except Exception as err:
            raise VisitException(self, err)

    def copy(self) -> SignalAssignment:
        return SignalAssignment(self._target, self._source)

    def update_transitions(self, state_map: IdMap):
        pass

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"SignalAssignment",
            content=[
                f"target = {Statement._obj_str(self._target)}",
                f"source = {Statement._obj_str(self._source)}",
            ],
        )


class SignalPush(Statement):
    def __init__(self, target_signal: Signal, source):
        self._target = target_signal
        self._source = source

        super().__init__()

    def visit_objects(self, operation: Callable):
        try:
            self._target = operation(self._target, AccessFlags.PUSH)
            self._source = operation(self._source, AccessFlags.READ)
        except Exception as err:
            raise VisitException(self, err)

    def copy(self) -> SignalPush:
        return SignalPush(self._target, self._source)

    def update_transitions(self, state_map: IdMap):
        pass

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"SignalPush",
            content=[
                f"target = {Statement._obj_str(self._target)}",
                f"source = {Statement._obj_str(self._source)}",
            ],
        )


class _SignalAlias(Statement):
    """
    _SignalAlias is used during parsing to indicate, that reads from
    a signal should instead refer to the given replacement.
    This allows signals to be used in the same state they are initialized.
    """

    def __init__(self, signal: Signal, replacement: Temporary):
        self.signal = signal
        self.replacement = replacement

    def visit_objects(self, operation: Callable): ...

    def copy(self) -> _SignalAlias:
        return _SignalAlias(self.signal, self.replacement)

    def update_transitions(self, state_map: IdMap):
        raise AssertionError("_SignalAlias should not be present after parsing")

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"_SignalAlias",
            content=[
                f"signal = {Statement._obj_str(self.signal)}",
                f"replacement = {Statement._obj_str(self.replacement)}",
            ],
        )


class VariableAssignment(Statement):
    def __init__(self, target_var, source):
        self._target = target_var
        self._source = source

        super().__init__()

    def visit_objects(self, operation: Callable):
        try:
            self._target = operation(self._target, AccessFlags.WRITE)
            self._source = operation(self._source, AccessFlags.READ)
        except Exception as err:
            raise VisitException(self, err)

    def copy(self) -> VariableAssignment:
        return VariableAssignment(self._target, self._source)

    def update_transitions(self, state_map: IdMap):
        pass

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="VariableAssignment",
            content=[
                f"target = {Statement._obj_str(self._target)}",
                f"source = {Statement._obj_str(self._source)}",
            ],
        )


class ResetInstance(Statement):
    """
    statement, that indicates, that a given synthesizable
    should be set to its default value
    """

    def __init__(self, obj):
        self._obj = obj

        super().__init__()

    def visit_objects(self, operation: Callable):
        try:
            self._obj = operation(self._obj, AccessFlags.WRITE)
        except Exception as err:
            raise VisitException(self, err)

    def copy(self) -> ResetInstance:
        return ResetInstance(self._obj)

    def update_transitions(self, state_map: IdMap):
        pass

    def dump(self) -> IndentBlock:
        return IndentBlock(title=f"ResetInstance ({self._obj})")


#
#
# intrinsic functions
#
#


class Boolean(Expression):
    def __init__(
        self,
        arg,
        result: Temporary[Bit],
    ):
        self._arg = arg
        super().__init__(result)

    def visit_objects(self, operation: Callable):
        try:
            self._arg = operation(self._arg, AccessFlags.READ)
            self._result = operation(self._result, AccessFlags.WRITE)
        except Exception as err:
            raise VisitException(self, err)

    def dump(self) -> str:
        return "Boolean"

    def copy(self) -> Boolean:
        return Boolean(self._arg, self._result)


class SelectWith(Expression):
    def __init__(
        self,
        arg,
        branches: list[Tuple[Any, Any]],
        default: Any | None,
        result,
    ):
        self._arg = arg
        self._branches = [[*branch] for branch in branches]
        self._default = default

        super().__init__(result)

    def visit_objects(self, operation: Callable):
        try:
            self._arg = operation(self._arg, AccessFlags.READ)

            self._branches = [
                [operation(a, AccessFlags.READ), operation(b, AccessFlags.READ)]
                for a, b in self._branches
            ]

            if self._default is not None:
                self._default = operation(self._default, AccessFlags.READ)

            self._result = operation(self._result, AccessFlags.WRITE)
        except Exception as err:
            raise VisitException(self, err)

    def dump(self) -> IndentBlock:
        if self._default is None:
            default = []
        else:
            default = [IndentBlock(title="[default]", content=f"{self._default}")]

        return IndentBlock(
            title="SelectWith",
            content=[
                IndentBlock(title=f"[{cond}]", content=f"{value}")
                for cond, value in self._branches
            ]
            + default,
        )

    def copy(self) -> SelectWith:
        return SelectWith(
            self._arg, [(a, b) for a, b in self._branches], self._default, self._result
        )


class CaseWhen(Statement):
    class Branch:
        def __init__(self, cond, body: CodeBlock):
            self.cond = cond
            self.code = body

        def __iter__(self):
            return iter((self.cond, self.code))

    def __init__(
        self, value, branches: list[Tuple], default: CodeBlock | None, frame=None
    ):
        self._value = value
        self._branches = [CaseWhen.Branch(cond, body) for cond, body in branches]
        self._default = default
        super().__init__()

        if frame is not None:
            self._frame = frame

    def visit(self, operation):
        try:
            for branch in self._branches:
                branch.code = branch.code.visit(operation)

            if self._default is not None:
                self._default = self._default.visit(operation)

            return operation(self)
        except Exception as err:
            raise VisitException(self, err)

    def visit_objects(self, operation: Callable):
        try:
            new_branches = []

            for branch in self._branches:
                cond = operation(branch.cond, AccessFlags.READ)
                branch.code.visit_objects(operation)

                new_branches.append(CaseWhen.Branch(cond, branch.code))

            self._branches = new_branches

            if self._default is not None:
                self._default.visit_objects(operation)

            self._value = operation(self._value, AccessFlags.READ)
        except Exception as err:
            raise VisitException(self, err)

    def dump(self) -> IndentBlock:
        if self._default is None:
            return IndentBlock(
                title="CaseWhen",
                content=[
                    IndentBlock([f"{branch.cond}", branch.code.dump()])
                    for branch in self._branches
                ],
            )
        else:
            return IndentBlock(
                title="CaseWhen",
                content=[
                    IndentBlock([f"{branch.cond}", branch.code.dump()])
                    for branch in self._branches
                ]
                + [IndentBlock(title="default", content=self._default.dump())],
            )

    def copy(self) -> CaseWhen:
        return CaseWhen(
            self._value,
            [(branch.cond, branch.code.copy()) for branch in self._branches],
            None if self._default is None else self._default.copy(),
        )

    def update_transitions(self, state_map: IdMap):
        for branch in self._branches:
            branch.code.update_transitions(state_map)

        if self._default is not None:
            self._default.update_transitions(state_map)


class CondSelect(Statement):
    def __init__(
        self, branches: list[Tuple[Expression, CodeBlock]], default: CodeBlock | None
    ):
        self._branches = branches
        self._default = default
        super().__init__()

    def visit(self, operation):
        try:
            self._branches = [
                (expr.visit(operation), branch.visit(operation))
                for expr, branch in self._branches
            ]

            if self._default is not None:
                self._default = self._default.visit(operation)

            return operation(self)
        except Exception as err:
            raise VisitException(self, err)

    def visit_objects(self, operation: Callable):
        try:
            for branch in self._branches:
                branch[0].visit_objects(operation)
                branch[1].visit_objects(operation)

            if self._default is not None:
                self._default.visit_objects(operation)
        except Exception as err:
            raise VisitException(self, err)

    def dump(self) -> IndentBlock:
        if self._default is None:
            return IndentBlock(
                title="CondSelect",
                content=[
                    IndentBlock(
                        [expr.dump(), code.dump()] for expr, code in self._branches
                    )
                ],
            )
        else:
            return IndentBlock(
                title="CondSelect",
                content=[
                    IndentBlock(
                        [expr.dump(), code.dump()] for expr, code in self._branches
                    )
                ]
                + [IndentBlock(title="default", content=self._default.dump())],
            )

    def copy(self) -> CondSelect:
        return CondSelect(
            [(expr.copy(), code.copy()) for expr, code in self._branches],
            None if self._default is None else self._default.copy(),
        )

    def update_transitions(self, state_map: IdMap):
        for branch in self._branches:
            branch[1].update_transitions(state_map)

        if self._default is not None:
            self._default.update_transitions(state_map)


class _State(Statement):
    def __init__(self, code: CodeBlock, open_block: CodeBlock):
        self._code = code
        self._state_id = None

        # a code block that is part of self._code
        # and used to add statments
        self._open_block = open_block

        code.set_parent_state(self)
        open_block.set_parent_state(self)

        # init without used objects because
        # code is not complete at this point
        # instead tracked_objects is overridden and calls
        # self._code.tracked_objects()
        super().__init__()

    def fix_alias(self):
        # replace all reads from locally constructed signals
        # with reads from a temporary

        self._code._fix_alias()

    def visit(self, operation):
        try:
            self._code = self._code.visit(operation)
            return operation(self)
        except Exception as err:
            raise VisitException(self, err)

    def visit_objects(self, operation: Callable):
        self._code.visit_objects(operation)

    def code(self):
        return self._code

    def set_state_id(self, id: cohdl_enum.Enum):
        self._state_id = id

    def get_state_id(self):
        return self._state_id

    def update_transitions(self, state_map: IdMap):
        self._code.update_transitions(state_map)

    def copy(self) -> _State:
        class DummyOpenBlock:
            """
            open_block is only required during parsing,
            when copying a complete State DummyOpenBlock is passed instead
            it should never be used
            """

            def set_parent_state(self, state):
                pass

        return _State(self._code.copy(), cast(CodeBlock, DummyOpenBlock()))

    def empty(self) -> bool:
        return self._code.empty()

    def open_block(self):
        return self._open_block

    def append(self, stmt: Statement):
        self._open_block.append(stmt)
        stmt.set_parent_state(self)

    def set_open_block(self, open_block: CodeBlock):
        self._open_block = open_block
        open_block.set_parent_state(self)

    def dump(self) -> IndentBlock:
        return IndentBlock(title=f"State, id: {id(self):#x}", content=self._code.dump())


class _Transition(Statement):
    def __init__(self, next_state: _State):
        super().__init__()
        self._next_state = next_state

    def visit_objects(self, operation: Callable):
        pass

    def copy(self) -> _Transition:
        return _Transition(self._next_state)

    def dump(self) -> IndentBlock:
        return IndentBlock(title=f"Transition, target={self._next_state}", content=[])


class _ResetContext(Statement):
    def visit_objects(self, operation: Callable):
        pass

    def copy(self):
        return _ResetContext()

    def dump(self) -> IndentBlock:
        return IndentBlock(title=f"ResetContext")


class _ResetPushed(Statement):
    def visit_objects(self, operation: Callable):
        pass

    def copy(self):
        return _ResetPushed()

    def dump(self) -> IndentBlock:
        return IndentBlock(title=f"ResetPushed", content=[])


class BitSignalEvent(Expression):
    Type = _BitSignalEvent.Type

    def __init__(self, event_type: BitSignalEvent.Type, arg, result):
        super().__init__(result)
        self.arg = arg
        self.event_type = event_type

    def visit_objects(self, operation: Callable):
        try:
            self.arg = operation(self.arg, AccessFlags.READ)
            self._result = operation(self._result, AccessFlags.WRITE)
        except Exception as err:
            raise VisitException(self, err)

    def copy(self):
        return BitSignalEvent(self.event_type, self.arg, self.result())

    def dump(self) -> IndentBlock:
        return IndentBlock(title=f"BitsignalEvent")


class StatemachineContext:
    _singleton: StatemachineContext | None = None

    @staticmethod
    def enter(name: str):
        assert (
            StatemachineContext._singleton is None
        ), "error nested StatemachineContext"
        ctx = StatemachineContext(name)
        StatemachineContext._singleton = ctx
        return ctx

    @staticmethod
    def get():
        ctx = StatemachineContext._singleton
        assert ctx is not None, "internal error: requested statemachine context"
        return ctx

    @staticmethod
    def finish(open_blocks):
        assert (
            StatemachineContext._singleton is not None
        ), "internal error: finished StatemachineContext with no active singleton"
        ctx = StatemachineContext._singleton
        StatemachineContext._singleton = None

        ctx._fix_signal_alias()

        if len(ctx._states) == 0:
            return CodeBlock([], None)
        elif len(ctx._states) == 1:
            # when the statemachine consists of only one state,
            # it can be treated as a normal code block
            # no transition logic is needed

            code = ctx._states[0].code()

            def remove_transitions(stmt):
                if isinstance(stmt, _Transition):
                    return Nop()
                return stmt

            code.visit(remove_transitions)

            return code

        # if a statemachine contains multiple states
        # add a transition back to the first state to all code blocks, that reach the end
        # of the sequential block
        # insert transition at the front of the block because it is possible, that
        # the final block already contains transitions (potentially in nested if statements)
        # that would be shadowed by a transition at the very end
        for open_block in open_blocks:
            open_block.addfront(_Transition(ctx.first_state()))

        return Statemachine(ctx)

    def _check_temporaries(self):
        for state in self._states:
            used_temporaries: IdMap[Any, bool] = IdMap()

            def check(obj, access: AccessFlags):
                if isinstance(obj, Temporary):
                    if obj._root not in used_temporaries:
                        # the first access to a temporary within a
                        # state must be a write access
                        # since writes to temporaries are only possible during
                        # construction this check also ensures, that
                        # no temporary is used (read) outside the single
                        # state it is valid in
                        assert (
                            access is AccessFlags.WRITE
                        ), "Temporary objects may not be shared between states"
                        used_temporaries[obj._root] = True
                return obj

            state.visit_objects(check)

    def _fix_signal_alias(self):
        for state in self._states:
            state.fix_alias()

    def _set_state_signal(self, signal):
        def set_signal(stmt):
            if isinstance(stmt, _Transition):
                stmt._state_signal = signal
            return stmt

        for state in self._states:
            state.visit(set_signal)

    def __init__(self, name: str | None = None):
        first_block = CodeBlock([], None)
        self._first = _State(first_block, first_block)
        self._states: list[_State] = [self._first]
        self._name = name

    def first_block(self):
        return self._first.code()

    def at_start(self):
        return self._first.empty()

    def first_state(self):
        return self._first

    def add_state(self, state: _State):
        self._states.append(state)


class Statemachine(Statement):
    def __init__(self, ctx: StatemachineContext):
        super().__init__()

        self._ctx = ctx

        name = ctx._name
        states = ctx._states

        if name is None:
            enum_name = "process_state"
        else:
            if name.startswith("_"):
                enum_name = f"state{name}"
            else:
                enum_name = f"state_{name}"

        self._state_type = cohdl_enum.Enum(
            enum_name, [f"state_{nr}" for nr in range(len(states))]
        )

        state_name = f"s{name}" if name.startswith("_") else f"s_{name}"

        self._current_state = Signal[self._state_type](
            self._state_type(1), name=state_name
        )

        ctx._set_state_signal(self._current_state)

        self._state_id = IdMap()

        for nr, state in enumerate(states):
            self._state_id[state] = self._state_type(nr + 1)

    def check_temporaries(self):
        self._ctx._check_temporaries()

    def as_case_when(self):
        def replace_transition(stmt):
            if isinstance(stmt, _Transition):
                return SignalAssignment(
                    self._current_state, self._state_id[stmt._next_state], stmt._frame
                )
            else:
                return stmt

        for state in self._ctx._states:
            state.visit(replace_transition)

        return CaseWhen(
            self._current_state,
            [(self._state_id[state], state.code()) for state in self._ctx._states],
            None,
            self._frame,
        )

    def visit(self, operation):
        try:
            self._ctx._states = [state.visit(operation) for state in self._ctx._states]
            return operation(self)
        except Exception as err:
            raise VisitException(self, err)

    def visit_objects(self, operation):
        try:
            self._current_state = operation(self._current_state, AccessFlags.READ)

            for state in self._ctx._states:
                state.visit_objects(operation)
        except Exception as err:
            raise VisitException(self, err)

    def dump(self):
        return IndentBlock("Statemachine")

    def copy(self):
        copied_states = [state.copy() for state in self._ctx._states]

        state_map = IdMap()

        for state, cpy_state in zip(self._ctx._states, copied_states):
            state_map[state] = (cpy_state, self)

        for cpy_state in copied_states:
            cpy_state.update_transitions(state_map)

    def update_transitions(self, state_map: IdMap):
        raise AssertionError("TODO")


#
# Assertions
#


class Assert(Statement):
    def __init__(self, cond, msg):
        self._cond = cond
        self._msg = msg
        super().__init__()

    def visit_objects(self, operation: Callable):
        try:
            self._cond = operation(self._cond, AccessFlags.READ)
        except Exception as err:
            raise VisitException(self, err)

    def update_transitions(self, state_map: IdMap):
        pass

    def copy(self) -> Assert:
        return Assert(self._cond, self._msg)

    def dump(self) -> IndentBlock:
        return IndentBlock(title=f"Assertation", content=str(self._msg))


#
# inline code
#


class InlineCode(Statement):
    def __init__(self, options: list[_InlineCode], result):
        self.options = options
        self.result = result

    def visit_objects(self, operation):
        try:

            def visit_nodes(nodes: list[_InlineCode.Node]):
                updated = []

                for node in nodes:
                    if isinstance(node, _InlineCode.Text):
                        updated.append(node)
                    elif isinstance(node, _InlineCode.Object):
                        access = AccessFlags.READ if node.read else AccessFlags.WRITE
                        updated.append(
                            _InlineCode.Object(operation(node.obj, access), node.read)
                        )
                    elif isinstance(node, _InlineCode.SubCode):
                        for option in node.options:
                            assert isinstance(option, _InlineCode)
                            visit_nodes(option.content)
                    else:
                        raise AssertionError("invalid content")

                return updated

            for option in self.options:
                visit_nodes(option.content)

            if self.result is not None:
                self.result = operation(self.result, AccessFlags.WRITE)
        except Exception as err:
            raise VisitException(self, err)

    def copy(self):
        return InlineCode([o.copy() for o in self.options])

    def dump(self):
        return IndentBlock(title="InlineCode")


#
#
# Instances
#
#


class Block:
    def __init__(
        self,
        name: str | None,
        subblocks: list[Block],
        contexts: list[Context],
        attributes: dict,
    ):
        self._name = name
        self._subblocks = subblocks
        self._contexts = contexts
        self._attributes = attributes

    def name(self):
        return self._name

    def subblocks(self) -> list[Block]:
        return self._subblocks

    def contexts(self):
        return self._contexts

    def all_contexts(self):
        # iterate over all nested contexts

        for ctx in self._contexts:
            yield ctx

        for sub in self._subblocks:
            yield from sub.all_contexts()

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="Block",
            content=[
                IndentBlock(
                    title="Subblocks", content=[inst.dump() for inst in self._subblocks]
                ),
                IndentBlock(
                    title="Contexts", content=[inst.dump() for inst in self._contexts]
                ),
            ],
        )


#
# synthesizable contexts
#


class Context:
    def __init__(
        self,
        name: str,
        code: CodeBlock,
        attributes: dict,
        source_location: SourceLocation,
    ):
        self._name = name
        self._code = code
        self.attributes = attributes
        self._source_location = source_location

    def visit(self, operation):
        try:
            self._code = self._code.visit(operation)
        except Exception as err:
            raise VisitException(self, err)

    def visit_objects(self, operation):
        self._code.visit_objects(operation)

    # TODO: maybe make this the default
    # and always visit all referenced objects
    visit_referenced_objects = _visit_referenced_objects

    def code(self):
        return self._code

    def name(self):
        return self._name

    def source_location(self):
        return self._source_location


class Concurrent(Context):
    def __init__(
        self,
        name: str,
        code: CodeBlock,
        attributes: dict,
        source_location: SourceLocation,
    ):
        super().__init__(name, code, attributes, source_location)

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="Concurrent",
            content=[
                IndentBlock(title="code", content=self._code.dump()),
            ],
        )


class Sequential(Context):
    def _pushed_resettable_signals(self):
        pushed = IdSet()
        resettable = IdSet()

        def visit_objects(obj, access: AccessFlags):
            if access & AccessFlags.PUSH:
                pushed.add(obj._root)

            if access & (AccessFlags.PUSH | AccessFlags.WRITE):
                root = obj._root
                if root.has_default() and not root._noreset:
                    resettable.add(root)

            return obj

        self.visit_objects(visit_objects)

        def visit_statements(stmt):
            if isinstance(stmt, _ResetContext):
                return CodeBlock(
                    [
                        (
                            SignalAssignment(r, r.default(), stmt._frame)
                            if isinstance(r, Signal)
                            else VariableAssignment(r, r.default())
                        )
                        for r in resettable
                    ],
                    None,
                )
            elif isinstance(stmt, _ResetPushed):
                return CodeBlock(
                    [
                        SignalAssignment(sig, sig.default(), stmt._frame)
                        for sig in pushed
                    ],
                    None,
                )

            return stmt

        self._code.visit(visit_statements)

    def __init__(
        self,
        name: str,
        code: CodeBlock,
        always_expr: Concurrent | None,
        sensitivity: _SensitivitySpec | None,
        attributes: dict,
        source_location: SourceLocation,
    ):
        if sensitivity is None:
            sensitivity = _SensitivityAll()

        assert isinstance(sensitivity, _SensitivitySpec)
        super().__init__(name, code, attributes, source_location)
        self._always_expr = always_expr
        self._sensitivity = sensitivity

        def translate_statemachine(stmt):
            if isinstance(stmt, Statemachine):
                stmt.check_temporaries()
                return stmt.as_case_when()
            else:
                return stmt

        code.visit(translate_statemachine)

        self._pushed_resettable_signals()
        code._fix_alias()

    def visit(self, operation):
        try:
            if self._always_expr is not None:
                self._always_expr._code = self._always_expr._code.visit(operation)
            super().visit(operation)
        except Exception as err:
            raise VisitException(self, err)

    def visit_objects(self, operation: Callable):
        try:
            if self._always_expr is not None:
                self._always_expr.code().visit_objects(operation)

            if isinstance(self._sensitivity, _SensitivityList):
                signals = self._sensitivity.signals
                signals = [operation(sig, AccessFlags.READ) for sig in signals]
            else:
                assert isinstance(self._sensitivity, _SensitivityAll)

            super().visit_objects(operation)
        except Exception as err:
            raise VisitException(self, err)

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="Sequential",
            content=[
                IndentBlock(title="code", content=self._code.dump()),
            ],
        )


class EntityTemplate(Block):
    """
    entity with unconstrained ports, generics
    """

    def __init__(
        self,
        info: EntityInfo,
        subblocks: list[Block],
        contexts: list[Context],
    ):
        super().__init__(info.name, subblocks, contexts, info.attributes)
        self._info = info

        #
        # check signal usage
        #

        written_in = IdMap()
        used_in = IdMap()

        current_ctx: Context

        def check_usage(obj, access: AccessFlags):
            # check, that signals are only written in a single context
            if access is AccessFlags.WRITE or access is AccessFlags.PUSH:
                if isinstance(obj, Port) and obj.is_input():
                    raise AssertionError(f"writing to input port '{obj}' not allowed")

                if isinstance(obj, (Signal, Variable, Temporary)):
                    obj_root = obj._root
                    if obj_root in written_in:
                        assert written_in[obj_root] is current_ctx, (
                            f"object '{obj_root}, name={obj_root.name()}' written in multiple contexts\n"
                            " written in this context\n"
                            f"{current_ctx.source_location()}\n"
                            " also written in this context\n"
                            f"{written_in[obj_root].source_location()}\n"
                        )
                    else:
                        written_in[obj_root] = current_ctx

            # check, that Variables and Temporaries are only used in a single context
            if isinstance(obj, (Temporary, Variable)):
                obj_root = obj._root
                if obj_root in used_in:
                    assert used_in[obj_root] is current_ctx, (
                        f"object '{obj_root}' used in multiple contexts\n"
                        " first found here\n"
                        f"{current_ctx.source_location()}\n"
                        " also found here\n"
                        f"{used_in[obj_root].source_location()}\n"
                    )
                else:
                    used_in[obj_root] = current_ctx

            return obj

        for ctx in self.all_contexts():
            current_ctx = ctx
            ctx.code().visit_objects(check_usage)

    def info(self):
        return self._info

    def extern(self):
        return self._info.extern

    def port_declarations(self) -> dict[str, Port]:
        return self._info.ports

    def generic_declarations(self) -> dict[str, Generic]:
        return self._info.generics

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"EntityTemplate {self._name}",
            content=[
                IndentBlock(
                    title="Ports",
                    content=[repr(port) for port in self._info.ports.values()],
                ),
                IndentBlock(
                    title="Subinstances",
                    content=[inst.dump() for inst in self._subblocks],
                ),
                IndentBlock(
                    title="Contexts",
                    content=[inst.dump() for inst in self._contexts],
                ),
            ],
        )


class Entity(Block):
    def __init__(
        self,
        template: EntityTemplate,
        name,
        ports: dict[str, Signal],
        generics: dict[str, Any],
    ):
        self._template: EntityTemplate = template

        self._ports: dict[str, Signal] = ports
        self._generics: dict[str, Any] = generics

        super().__init__(name, [], [], {})

    def get_ports(self):
        return self._ports

    def get_generics(self):
        return self._generics

    def get_template(self) -> EntityTemplate:
        return self._template

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"Entity ({self._name})",
            content=[
                IndentBlock(title="template", content=self._template.dump()),
                IndentBlock(
                    title="ports",
                    content=[f"{name}, {port}" for name, port in self._ports.items()],
                ),
            ],
        )



from cohdl import Signal, Bit, pyeval
from cohdl.std._context import sequential

from ._builtins import vhdl, n_prev_true
from ._sequence import _SeqNode, When


class Checker:
    def __init__(self, clk, reset=None, *, prefix: str | None = None):
        self._always_false = Signal[Bit](False, name="always_false")
        self._past_valid = Signal[Bit](False, name="past_exists")
        self._past_valid_after_reset = Signal[Bit](False, name="past_valid_after_reset")
        self._clk = clk
        self._reset = reset
        self._prefix = prefix

        @sequential(clk, reset)
        def proc():
            self._past_valid_after_reset <<= True

        @sequential(clk)
        def proc():
            self._past_valid <<= True

    @pyeval
    def _complete_label(self, label):
        if self._prefix is not None:
            return f"{self._prefix}{label}"
        return label

    def valid_since_reset(self, cnt: int | tuple = 0):
        return n_prev_true(self._past_valid_after_reset, cnt)

    def valid_since_start(self, cnt: int | tuple = 0):
        return n_prev_true(self._past_valid, cnt)

    def always(
        self,
        label,
        cond,
        *,
        since_start=None,
        since_reset=None,
    ):
        lbl = self._complete_label(label)

        if since_start is None:
            start_guard = True
        else:
            start_guard = self.valid_since_start(since_start)

        if since_reset is None:
            reset_guard = True
        else:
            reset_guard = self.valid_since_reset(since_reset)

        if start_guard is True and reset_guard is True:
            f"{vhdl:{lbl} : assert always {_SeqNode.write_node(cond)};}"
        else:
            f"{vhdl:{lbl} : assert always {_SeqNode.write_node(When(start_guard and reset_guard, cond))};}"

    #
    #
    #

    def assume_always(
        self,
        label,
        cond,
        *,
        since_start=None,
        since_reset=None,
    ):
        lbl = self._complete_label(label)

        if since_start is None:
            start_guard = True
        else:
            start_guard = self.valid_since_start(since_start)

        if since_reset is None:
            reset_guard = True
        else:
            reset_guard = self.valid_since_reset(since_reset)

        if start_guard is True and reset_guard is True:
            f"{vhdl:{lbl} : assume always {_SeqNode.write_node(cond)};}"
        else:
            f"{vhdl:{label} : assume always {_SeqNode.write_node(When(start_guard and reset_guard, cond))};}"

    def assume_initial(self, label, cond):
        f"{vhdl:{self._complete_label(label)} : assume {_SeqNode.write_node(cond)};}"

    #
    #
    #

    def cover(self, label, cond):
        f"{vhdl:{self._complete_label(label)} : cover <% {_SeqNode.write_node(cond)} %>;}"

    #
    #
    #

    def dbg_show(self, label, signal):
        self.always(label, signal, valid_for=10)

    def fail_after(self, label, cnt):
        self.always(
            label,
            When(self.valid_since_start(cnt - 1), self._always_false),
        )
